from __future__ import annotations

import uuid
from collections.abc import Collection
from datetime import datetime
from typing import cast

import sqlalchemy as sa

from ai.backend.common.data.filter_specs import StringMatchSpec, UUIDEqualMatchSpec, UUIDInMatchSpec
from ai.backend.common.types import KernelId, SessionId
from ai.backend.manager.data.deployment.types import RouteStatus
from ai.backend.manager.data.kernel.types import KernelSchedulingPhase
from ai.backend.manager.data.session.types import SchedulingResult, SessionStatus
from ai.backend.manager.models.scheduling_history import (
    DeploymentHistoryRow,
    KernelSchedulingHistoryRow,
    RouteHistoryRow,
    SessionSchedulingHistoryRow,
)
from ai.backend.manager.repositories.base import QueryCondition, QueryOrder

# ========== Session Scheduling History ==========


class SessionSchedulingHistoryConditions:
    """Query conditions for session scheduling history."""

    # UUID filter conditions for history id
    @staticmethod
    def by_id_filter(spec: UUIDEqualMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return SessionSchedulingHistoryRow.id != spec.value
            return SessionSchedulingHistoryRow.id == spec.value

        return inner

    @staticmethod
    def by_id_in(spec: UUIDInMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return SessionSchedulingHistoryRow.id.notin_(spec.values)
            return SessionSchedulingHistoryRow.id.in_(spec.values)

        return inner

    @staticmethod
    def by_ids(ids: Collection[uuid.UUID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.id.in_(ids)

        return inner

    @staticmethod
    def by_session_id(session_id: SessionId) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.session_id == session_id

        return inner

    @staticmethod
    def by_result(result: SchedulingResult) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.result == str(result)

        return inner

    @staticmethod
    def by_results(results: list[SchedulingResult]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.result.in_([str(r) for r in results])

        return inner

    @staticmethod
    def by_from_status(status: SessionStatus) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.from_status == str(status)

        return inner

    @staticmethod
    def by_from_statuses(statuses: list[str]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.from_status.in_(statuses)

        return inner

    @staticmethod
    def by_to_status(status: SessionStatus) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.to_status == str(status)

        return inner

    @staticmethod
    def by_to_statuses(statuses: list[str]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.to_status.in_(statuses)

        return inner

    @staticmethod
    def by_error_code(error_code: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.error_code == error_code

        return inner

    # UUID filter conditions for session_id
    @staticmethod
    def by_session_id_filter(spec: UUIDEqualMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return SessionSchedulingHistoryRow.session_id != spec.value
            return SessionSchedulingHistoryRow.session_id == spec.value

        return inner

    @staticmethod
    def by_session_id_in(spec: UUIDInMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return SessionSchedulingHistoryRow.session_id.notin_(spec.values)
            return SessionSchedulingHistoryRow.session_id.in_(spec.values)

        return inner

    # String filter conditions for error_code
    @staticmethod
    def by_error_code_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], SessionSchedulingHistoryRow.error_code)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}%"
            else:
                pattern = f"%{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_error_code_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], SessionSchedulingHistoryRow.error_code)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                val = spec.value.lower()
            else:
                val = spec.value
            if spec.negated:
                return col != val
            return col == val

        return inner

    @staticmethod
    def by_error_code_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], SessionSchedulingHistoryRow.error_code)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"{spec.value.lower()}%"
            else:
                pattern = f"{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_error_code_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], SessionSchedulingHistoryRow.error_code)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}"
            else:
                pattern = f"%{spec.value}"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    # String filter conditions for phase
    @staticmethod
    def by_phase_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str], SessionSchedulingHistoryRow.phase)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}%"
            else:
                pattern = f"%{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_phase_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str], SessionSchedulingHistoryRow.phase)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                val = spec.value.lower()
            else:
                val = spec.value
            if spec.negated:
                return col != val
            return col == val

        return inner

    @staticmethod
    def by_phase_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str], SessionSchedulingHistoryRow.phase)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"{spec.value.lower()}%"
            else:
                pattern = f"{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_phase_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str], SessionSchedulingHistoryRow.phase)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}"
            else:
                pattern = f"%{spec.value}"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    # String filter conditions for message
    @staticmethod
    def by_message_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], SessionSchedulingHistoryRow.message)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}%"
            else:
                pattern = f"%{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_message_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], SessionSchedulingHistoryRow.message)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                val = spec.value.lower()
            else:
                val = spec.value
            if spec.negated:
                return col != val
            return col == val

        return inner

    @staticmethod
    def by_message_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], SessionSchedulingHistoryRow.message)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"{spec.value.lower()}%"
            else:
                pattern = f"{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_message_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], SessionSchedulingHistoryRow.message)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}"
            else:
                pattern = f"%{spec.value}"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_cursor_forward(cursor_id: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor)."""
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(SessionSchedulingHistoryRow.created_at)
                .where(SessionSchedulingHistoryRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return SessionSchedulingHistoryRow.created_at < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_id: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor)."""
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(SessionSchedulingHistoryRow.created_at)
                .where(SessionSchedulingHistoryRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return SessionSchedulingHistoryRow.created_at > subquery

        return inner

    # DateTime filter conditions
    @staticmethod
    def by_created_at_before(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.created_at < dt

        return inner

    @staticmethod
    def by_created_at_after(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.created_at > dt

        return inner

    @staticmethod
    def by_created_at_equals(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.created_at == dt

        return inner

    @staticmethod
    def by_updated_at_before(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.updated_at < dt

        return inner

    @staticmethod
    def by_updated_at_after(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.updated_at > dt

        return inner

    @staticmethod
    def by_updated_at_equals(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.updated_at == dt

        return inner


class SessionSchedulingHistoryOrders:
    """Query orders for session scheduling history."""

    @staticmethod
    def created_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return SessionSchedulingHistoryRow.created_at.asc()
        return SessionSchedulingHistoryRow.created_at.desc()

    @staticmethod
    def updated_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return SessionSchedulingHistoryRow.updated_at.asc()
        return SessionSchedulingHistoryRow.updated_at.desc()


# ========== Kernel Scheduling History ==========


class KernelSchedulingHistoryConditions:
    """Query conditions for kernel scheduling history."""

    @staticmethod
    def by_kernel_id(kernel_id: KernelId) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return KernelSchedulingHistoryRow.kernel_id == kernel_id

        return inner

    @staticmethod
    def by_session_id(session_id: SessionId) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return KernelSchedulingHistoryRow.session_id == session_id

        return inner

    @staticmethod
    def by_result(result: SchedulingResult) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return KernelSchedulingHistoryRow.result == str(result)

        return inner

    @staticmethod
    def by_from_phase(phase: KernelSchedulingPhase) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return cast(
                sa.sql.expression.ColumnElement[bool],
                KernelSchedulingHistoryRow.from_phase == str(phase),
            )

        return inner

    @staticmethod
    def by_to_phase(phase: KernelSchedulingPhase) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return cast(
                sa.sql.expression.ColumnElement[bool],
                KernelSchedulingHistoryRow.to_phase == str(phase),
            )

        return inner

    @staticmethod
    def by_error_code(error_code: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return KernelSchedulingHistoryRow.error_code == error_code

        return inner

    @staticmethod
    def by_cursor_forward(cursor_id: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor)."""
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(KernelSchedulingHistoryRow.created_at)
                .where(KernelSchedulingHistoryRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return KernelSchedulingHistoryRow.created_at < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_id: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor)."""
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(KernelSchedulingHistoryRow.created_at)
                .where(KernelSchedulingHistoryRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return KernelSchedulingHistoryRow.created_at > subquery

        return inner


class KernelSchedulingHistoryOrders:
    """Query orders for kernel scheduling history."""

    @staticmethod
    def created_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return KernelSchedulingHistoryRow.created_at.asc()
        return KernelSchedulingHistoryRow.created_at.desc()

    @staticmethod
    def updated_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return KernelSchedulingHistoryRow.updated_at.asc()
        return KernelSchedulingHistoryRow.updated_at.desc()


# ========== Deployment History ==========


class DeploymentHistoryConditions:
    """Query conditions for deployment history."""

    # UUID filter conditions for history id
    @staticmethod
    def by_id_filter(spec: UUIDEqualMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return DeploymentHistoryRow.id != spec.value
            return DeploymentHistoryRow.id == spec.value

        return inner

    @staticmethod
    def by_id_in(spec: UUIDInMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return DeploymentHistoryRow.id.notin_(spec.values)
            return DeploymentHistoryRow.id.in_(spec.values)

        return inner

    @staticmethod
    def by_ids(ids: Collection[uuid.UUID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.id.in_(ids)

        return inner

    @staticmethod
    def by_deployment_id(deployment_id: uuid.UUID) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.deployment_id == deployment_id

        return inner

    @staticmethod
    def by_result(result: SchedulingResult) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.result == str(result)

        return inner

    @staticmethod
    def by_results(results: list[SchedulingResult]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.result.in_([str(r) for r in results])

        return inner

    @staticmethod
    def by_error_code(error_code: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.error_code == error_code

        return inner

    # UUID filter conditions for deployment_id
    @staticmethod
    def by_deployment_id_filter(spec: UUIDEqualMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return DeploymentHistoryRow.deployment_id != spec.value
            return DeploymentHistoryRow.deployment_id == spec.value

        return inner

    @staticmethod
    def by_deployment_id_in(spec: UUIDInMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return DeploymentHistoryRow.deployment_id.notin_(spec.values)
            return DeploymentHistoryRow.deployment_id.in_(spec.values)

        return inner

    # String filter conditions for error_code
    @staticmethod
    def by_error_code_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], DeploymentHistoryRow.error_code)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}%"
            else:
                pattern = f"%{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_error_code_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], DeploymentHistoryRow.error_code)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                val = spec.value.lower()
            else:
                val = spec.value
            if spec.negated:
                return col != val
            return col == val

        return inner

    @staticmethod
    def by_error_code_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], DeploymentHistoryRow.error_code)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"{spec.value.lower()}%"
            else:
                pattern = f"{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_error_code_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], DeploymentHistoryRow.error_code)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}"
            else:
                pattern = f"%{spec.value}"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    # Status list filter conditions
    @staticmethod
    def by_from_statuses(statuses: list[str]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.from_status.in_(statuses)

        return inner

    @staticmethod
    def by_to_statuses(statuses: list[str]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.to_status.in_(statuses)

        return inner

    # String filter conditions for phase
    @staticmethod
    def by_phase_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str], DeploymentHistoryRow.phase)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}%"
            else:
                pattern = f"%{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_phase_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str], DeploymentHistoryRow.phase)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                val = spec.value.lower()
            else:
                val = spec.value
            if spec.negated:
                return col != val
            return col == val

        return inner

    @staticmethod
    def by_phase_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str], DeploymentHistoryRow.phase)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"{spec.value.lower()}%"
            else:
                pattern = f"{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_phase_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str], DeploymentHistoryRow.phase)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}"
            else:
                pattern = f"%{spec.value}"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    # String filter conditions for message
    @staticmethod
    def by_message_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], DeploymentHistoryRow.message)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}%"
            else:
                pattern = f"%{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_message_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], DeploymentHistoryRow.message)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                val = spec.value.lower()
            else:
                val = spec.value
            if spec.negated:
                return col != val
            return col == val

        return inner

    @staticmethod
    def by_message_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], DeploymentHistoryRow.message)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"{spec.value.lower()}%"
            else:
                pattern = f"{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_message_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], DeploymentHistoryRow.message)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}"
            else:
                pattern = f"%{spec.value}"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_cursor_forward(cursor_id: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor)."""
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(DeploymentHistoryRow.created_at)
                .where(DeploymentHistoryRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return DeploymentHistoryRow.created_at < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_id: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor)."""
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(DeploymentHistoryRow.created_at)
                .where(DeploymentHistoryRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return DeploymentHistoryRow.created_at > subquery

        return inner

    # DateTime filter conditions
    @staticmethod
    def by_created_at_before(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.created_at < dt

        return inner

    @staticmethod
    def by_created_at_after(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.created_at > dt

        return inner

    @staticmethod
    def by_created_at_equals(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.created_at == dt

        return inner

    @staticmethod
    def by_updated_at_before(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.updated_at < dt

        return inner

    @staticmethod
    def by_updated_at_after(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.updated_at > dt

        return inner

    @staticmethod
    def by_updated_at_equals(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.updated_at == dt

        return inner


class DeploymentHistoryOrders:
    """Query orders for deployment history."""

    @staticmethod
    def created_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return DeploymentHistoryRow.created_at.asc()
        return DeploymentHistoryRow.created_at.desc()

    @staticmethod
    def updated_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return DeploymentHistoryRow.updated_at.asc()
        return DeploymentHistoryRow.updated_at.desc()


# ========== Route History ==========


class RouteHistoryConditions:
    """Query conditions for route history."""

    # UUID filter conditions for history id
    @staticmethod
    def by_id_filter(spec: UUIDEqualMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return RouteHistoryRow.id != spec.value
            return RouteHistoryRow.id == spec.value

        return inner

    @staticmethod
    def by_id_in(spec: UUIDInMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return RouteHistoryRow.id.notin_(spec.values)
            return RouteHistoryRow.id.in_(spec.values)

        return inner

    @staticmethod
    def by_ids(ids: Collection[uuid.UUID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.id.in_(ids)

        return inner

    @staticmethod
    def by_route_id(route_id: uuid.UUID) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.route_id == route_id

        return inner

    @staticmethod
    def by_deployment_id(deployment_id: uuid.UUID) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.deployment_id == deployment_id

        return inner

    @staticmethod
    def by_result(result: SchedulingResult) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.result == str(result)

        return inner

    @staticmethod
    def by_results(results: list[SchedulingResult]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.result.in_([str(r) for r in results])

        return inner

    @staticmethod
    def by_from_status(status: RouteStatus) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return cast(
                sa.sql.expression.ColumnElement[bool], RouteHistoryRow.from_status == status.value
            )

        return inner

    @staticmethod
    def by_to_status(status: RouteStatus) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return cast(
                sa.sql.expression.ColumnElement[bool], RouteHistoryRow.to_status == status.value
            )

        return inner

    @staticmethod
    def by_error_code(error_code: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.error_code == error_code

        return inner

    # UUID filter conditions for route_id
    @staticmethod
    def by_route_id_filter(spec: UUIDEqualMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return RouteHistoryRow.route_id != spec.value
            return RouteHistoryRow.route_id == spec.value

        return inner

    @staticmethod
    def by_route_id_in(spec: UUIDInMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return RouteHistoryRow.route_id.notin_(spec.values)
            return RouteHistoryRow.route_id.in_(spec.values)

        return inner

    # UUID filter conditions for deployment_id
    @staticmethod
    def by_deployment_id_filter(spec: UUIDEqualMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return RouteHistoryRow.deployment_id != spec.value
            return RouteHistoryRow.deployment_id == spec.value

        return inner

    @staticmethod
    def by_deployment_id_in(spec: UUIDInMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return RouteHistoryRow.deployment_id.notin_(spec.values)
            return RouteHistoryRow.deployment_id.in_(spec.values)

        return inner

    # String filter conditions for error_code
    @staticmethod
    def by_error_code_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], RouteHistoryRow.error_code)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}%"
            else:
                pattern = f"%{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_error_code_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], RouteHistoryRow.error_code)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                val = spec.value.lower()
            else:
                val = spec.value
            if spec.negated:
                return col != val
            return col == val

        return inner

    @staticmethod
    def by_error_code_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], RouteHistoryRow.error_code)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"{spec.value.lower()}%"
            else:
                pattern = f"{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_error_code_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], RouteHistoryRow.error_code)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}"
            else:
                pattern = f"%{spec.value}"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    # Status list filter conditions
    @staticmethod
    def by_from_statuses(statuses: list[str]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.from_status.in_(statuses)

        return inner

    @staticmethod
    def by_to_statuses(statuses: list[str]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.to_status.in_(statuses)

        return inner

    # String filter conditions for phase
    @staticmethod
    def by_phase_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str], RouteHistoryRow.phase)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}%"
            else:
                pattern = f"%{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_phase_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str], RouteHistoryRow.phase)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                val = spec.value.lower()
            else:
                val = spec.value
            if spec.negated:
                return col != val
            return col == val

        return inner

    @staticmethod
    def by_phase_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str], RouteHistoryRow.phase)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"{spec.value.lower()}%"
            else:
                pattern = f"{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_phase_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str], RouteHistoryRow.phase)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}"
            else:
                pattern = f"%{spec.value}"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    # String filter conditions for message
    @staticmethod
    def by_message_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], RouteHistoryRow.message)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}%"
            else:
                pattern = f"%{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_message_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], RouteHistoryRow.message)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                val = spec.value.lower()
            else:
                val = spec.value
            if spec.negated:
                return col != val
            return col == val

        return inner

    @staticmethod
    def by_message_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], RouteHistoryRow.message)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"{spec.value.lower()}%"
            else:
                pattern = f"{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_message_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], RouteHistoryRow.message)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}"
            else:
                pattern = f"%{spec.value}"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_cursor_forward(cursor_id: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor)."""
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(RouteHistoryRow.created_at)
                .where(RouteHistoryRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return RouteHistoryRow.created_at < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_id: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor)."""
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(RouteHistoryRow.created_at)
                .where(RouteHistoryRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return RouteHistoryRow.created_at > subquery

        return inner

    # DateTime filter conditions
    @staticmethod
    def by_created_at_before(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.created_at < dt

        return inner

    @staticmethod
    def by_created_at_after(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.created_at > dt

        return inner

    @staticmethod
    def by_created_at_equals(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.created_at == dt

        return inner

    @staticmethod
    def by_updated_at_before(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.updated_at < dt

        return inner

    @staticmethod
    def by_updated_at_after(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.updated_at > dt

        return inner

    @staticmethod
    def by_updated_at_equals(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.updated_at == dt

        return inner


class RouteHistoryOrders:
    """Query orders for route history."""

    @staticmethod
    def created_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return RouteHistoryRow.created_at.asc()
        return RouteHistoryRow.created_at.desc()

    @staticmethod
    def updated_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return RouteHistoryRow.updated_at.asc()
        return RouteHistoryRow.updated_at.desc()



from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.services.vfs_storage.actions.base import VFSStorageAction


@dataclass
class GetQuotaScopeAction(VFSStorageAction):
    storage_host_name: str
    quota_scope_id: str

    @override
    def entity_id(self) -> str | None:
        return self.quota_scope_id

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.GET


@dataclass
class GetQuotaScopeActionResult(BaseActionResult):
    quota_scope_id: str
    storage_host_name: str
    usage_bytes: int | None
    usage_count: int | None
    hard_limit_bytes: int | None

    @override
    def entity_id(self) -> str | None:
        return self.quota_scope_id



"""RBAC GQL types package."""

from .entity import (
    EntityConnection,
    EntityEdge,
    EntityFilter,
    EntityOrderBy,
    EntityOrderField,
    EntityRefGQL,
)
from .entity_node import EntityNode
from .permission import (
    CreatePermissionInput,
    DeletePermissionInput,
    DeletePermissionPayload,
    OperationTypeGQL,
    PermissionConnection,
    PermissionEdge,
    PermissionFilter,
    PermissionGQL,
    PermissionOrderBy,
    PermissionOrderField,
    RBACElementTypeGQL,
    ScopeEntityCombinationGQL,
    UpdatePermissionInput,
)
from .role import (
    AssignRoleInput,
    BulkAssignRoleErrorGQL,
    BulkAssignRoleInputGQL,
    BulkAssignRolePayloadGQL,
    BulkRevokeRoleErrorGQL,
    BulkRevokeRoleInputGQL,
    BulkRevokeRolePayloadGQL,
    CreateRoleInput,
    DeleteRoleInput,
    DeleteRolePayload,
    PurgeRoleInput,
    PurgeRolePayload,
    RevokeRoleInput,
    RoleAssignmentConnection,
    RoleAssignmentEdge,
    RoleAssignmentFilter,
    RoleAssignmentGQL,
    RoleAssignmentOrderBy,
    RoleAssignmentOrderField,
    RoleAssignmentRoleNestedFilterGQL,
    RoleConnection,
    RoleEdge,
    RoleFilter,
    RoleGQL,
    RoleOrderBy,
    RoleOrderField,
    RoleSourceGQL,
    RoleStatusGQL,
    UpdateRoleInput,
)

__all__ = [
    # Permission enums
    "RBACElementTypeGQL",
    "OperationTypeGQL",
    "PermissionOrderField",
    # Role enums
    "RoleSourceGQL",
    "RoleStatusGQL",
    "RoleOrderField",
    "RoleAssignmentOrderField",
    # Entity enums
    "EntityOrderField",
    # Types
    "PermissionGQL",
    "RoleGQL",
    "RoleAssignmentGQL",
    "EntityRefGQL",
    # Filters
    "PermissionFilter",
    "RoleFilter",
    "RoleAssignmentFilter",
    "RoleAssignmentRoleNestedFilterGQL",
    "EntityFilter",
    # OrderBy
    "PermissionOrderBy",
    "RoleOrderBy",
    "RoleAssignmentOrderBy",
    "EntityOrderBy",
    # Inputs
    "CreatePermissionInput",
    "UpdatePermissionInput",
    "DeletePermissionInput",
    "CreateRoleInput",
    "UpdateRoleInput",
    "DeleteRoleInput",
    "PurgeRoleInput",
    "AssignRoleInput",
    "RevokeRoleInput",
    "BulkAssignRoleInputGQL",
    "BulkRevokeRoleInputGQL",
    # Payloads
    "DeletePermissionPayload",
    "DeleteRolePayload",
    "PurgeRolePayload",
    "BulkAssignRoleErrorGQL",
    "BulkAssignRolePayloadGQL",
    "BulkRevokeRoleErrorGQL",
    "BulkRevokeRolePayloadGQL",
    # Connections
    "PermissionConnection",
    "PermissionEdge",
    "RoleConnection",
    "RoleEdge",
    "RoleAssignmentConnection",
    "RoleAssignmentEdge",
    # Entity types
    "EntityNode",
    "EntityEdge",
    "EntityConnection",
    # Scope-entity combination
    "ScopeEntityCombinationGQL",
]



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Final

from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.container_registry import get_container_registry_cls
from ai.backend.manager.container_registry.harbor import HarborRegistry_v2
from ai.backend.manager.data.container_registry.types import ContainerRegistryData
from ai.backend.manager.errors.image import (
    ContainerRegistryNotFound,
    ContainerRegistryWebhookAuthorizationFailed,
    HarborWebhookContainerRegistryRowNotFound,
)
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.repositories.container_registry.repository import (
    ContainerRegistryRepository,
)
from ai.backend.manager.services.container_registry.actions.clear_images import (
    ClearImagesAction,
    ClearImagesActionResult,
)
from ai.backend.manager.services.container_registry.actions.create_container_registry import (
    CreateContainerRegistryAction,
    CreateContainerRegistryActionResult,
)
from ai.backend.manager.services.container_registry.actions.create_registry_quota import (
    CreateRegistryQuotaAction,
    CreateRegistryQuotaActionResult,
)
from ai.backend.manager.services.container_registry.actions.delete_container_registry import (
    DeleteContainerRegistryAction,
    DeleteContainerRegistryActionResult,
)
from ai.backend.manager.services.container_registry.actions.delete_registry_quota import (
    DeleteRegistryQuotaAction,
    DeleteRegistryQuotaActionResult,
)
from ai.backend.manager.services.container_registry.actions.get_container_registries import (
    GetContainerRegistriesAction,
    GetContainerRegistriesActionResult,
)
from ai.backend.manager.services.container_registry.actions.handle_harbor_webhook import (
    HandleHarborWebhookAction,
    HandleHarborWebhookActionResult,
)
from ai.backend.manager.services.container_registry.actions.load_all_container_registries import (
    LoadAllContainerRegistriesAction,
    LoadAllContainerRegistriesActionResult,
)
from ai.backend.manager.services.container_registry.actions.load_container_registries import (
    LoadContainerRegistriesAction,
    LoadContainerRegistriesActionResult,
)
from ai.backend.manager.services.container_registry.actions.modify_container_registry import (
    ModifyContainerRegistryAction,
    ModifyContainerRegistryActionResult,
)
from ai.backend.manager.services.container_registry.actions.read_registry_quota import (
    ReadRegistryQuotaAction,
    ReadRegistryQuotaActionResult,
)
from ai.backend.manager.services.container_registry.actions.rescan_images import (
    RescanImagesAction,
    RescanImagesActionResult,
)
from ai.backend.manager.services.container_registry.actions.search_container_registries import (
    SearchContainerRegistriesAction,
    SearchContainerRegistriesActionResult,
)
from ai.backend.manager.services.container_registry.actions.update_registry_quota import (
    UpdateRegistryQuotaAction,
    UpdateRegistryQuotaActionResult,
)

if TYPE_CHECKING:
    from ai.backend.manager.models.container_registry import ContainerRegistryRow
    from ai.backend.manager.service.container_registry.harbor import (
        AbstractPerProjectContainerRegistryQuotaService,
    )

log: Final = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ContainerRegistryService:
    _db: ExtendedAsyncSAEngine
    _container_registry_repository: ContainerRegistryRepository
    _quota_service: AbstractPerProjectContainerRegistryQuotaService | None

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
        container_registry_repository: ContainerRegistryRepository,
        quota_service: AbstractPerProjectContainerRegistryQuotaService | None = None,
    ) -> None:
        self._db = db
        self._container_registry_repository = container_registry_repository
        self._quota_service = quota_service

    async def create_container_registry(
        self, action: CreateContainerRegistryAction
    ) -> CreateContainerRegistryActionResult:
        data = await self._container_registry_repository.create_registry(action.creator)
        return CreateContainerRegistryActionResult(data=data)

    async def modify_container_registry(
        self, action: ModifyContainerRegistryAction
    ) -> ModifyContainerRegistryActionResult:
        data = await self._container_registry_repository.modify_registry(action.updater)
        return ModifyContainerRegistryActionResult(data=data)

    async def delete_container_registry(
        self, action: DeleteContainerRegistryAction
    ) -> DeleteContainerRegistryActionResult:
        data = await self._container_registry_repository.delete_registry(action.purger)
        return DeleteContainerRegistryActionResult(data=data)

    async def rescan_images(self, action: RescanImagesAction) -> RescanImagesActionResult:
        registry_name = action.registry
        project = action.project

        registry_row: ContainerRegistryRow = (
            await self._container_registry_repository.get_registry_row_for_scanner(
                registry_name, project
            )
        )

        scanner_cls = get_container_registry_cls(registry_row)
        scanner = scanner_cls(self._db, registry_name, registry_row)
        result = await scanner.rescan_single_registry(action.progress_reporter)

        return RescanImagesActionResult(
            images=result.images, errors=result.errors, registry=registry_row.to_dataclass()
        )

    async def clear_images(self, action: ClearImagesAction) -> ClearImagesActionResult:
        registry_data = await self._container_registry_repository.clear_images(
            action.registry, action.project
        )

        return ClearImagesActionResult(registry=registry_data)

    async def load_container_registries(
        self, action: LoadContainerRegistriesAction
    ) -> LoadContainerRegistriesActionResult:
        registries: list[ContainerRegistryData] = []
        if action.project is not None:
            try:
                registry_data = (
                    await self._container_registry_repository.get_by_registry_and_project(
                        action.registry, action.project
                    )
                )
                registries = [registry_data]
            except ContainerRegistryNotFound:
                registries = []
        else:
            registries = await self._container_registry_repository.get_by_registry_name(
                action.registry
            )

        return LoadContainerRegistriesActionResult(registries=registries)

    async def load_all_container_registries(
        self, _action: LoadAllContainerRegistriesAction
    ) -> LoadAllContainerRegistriesActionResult:
        registries = await self._container_registry_repository.get_all()
        return LoadAllContainerRegistriesActionResult(registries=registries)

    async def search_container_registries(
        self, action: SearchContainerRegistriesAction
    ) -> SearchContainerRegistriesActionResult:
        """Search container registries with pagination and ordering."""
        result = await self._container_registry_repository.search_container_registries(
            action.querier
        )
        return SearchContainerRegistriesActionResult(
            data=result.items,
            total_count=result.total_count,
            has_next_page=result.has_next_page,
            has_previous_page=result.has_previous_page,
        )

    async def get_container_registries(
        self, _action: GetContainerRegistriesAction
    ) -> GetContainerRegistriesActionResult:
        registries = await self._container_registry_repository.get_known_registries()
        return GetContainerRegistriesActionResult(registries=registries)

    async def handle_harbor_webhook(
        self, action: HandleHarborWebhookAction
    ) -> HandleHarborWebhookActionResult:
        """Handle a Harbor container registry webhook event."""
        for resource in action.resources:
            registry_url = resource.resource_url.split("/")[0]
            registry_row = (
                await self._container_registry_repository.get_registry_by_url_and_project(
                    registry_url, action.project
                )
            )
            if not registry_row:
                raise HarborWebhookContainerRegistryRowNotFound(
                    extra_msg=(
                        f"Harbor webhook triggered, but the matching container registry"
                        f" row not found! (registry_url: {registry_url},"
                        f" project: {action.project})"
                    ),
                )

            # Validate webhook authorization
            if action.auth_header:
                extra = registry_row.extra or {}
                if extra.get("webhook_auth_header") != action.auth_header:
                    raise ContainerRegistryWebhookAuthorizationFailed(
                        extra_msg=(
                            f"Unauthorized webhook request"
                            f" (registry: {registry_row.registry_name},"
                            f" project: {action.project})"
                        ),
                    )

            # Handle event by type
            if action.event_type == "PUSH_ARTIFACT":
                scanner = HarborRegistry_v2(self._db, registry_row.registry_name, registry_row)
                await scanner.scan_single_ref(f"{action.project}/{action.img_name}:{resource.tag}")
            else:
                log.debug(
                    'Ignore harbor webhook event: "{}". Recommended to modify the'
                    " webhook config to not subscribe to this event type.",
                    action.event_type,
                )

        return HandleHarborWebhookActionResult()

    def _ensure_quota_service(self) -> AbstractPerProjectContainerRegistryQuotaService:
        if self._quota_service is None:
            raise RuntimeError("Registry quota service is not configured")
        return self._quota_service

    async def create_registry_quota(
        self, action: CreateRegistryQuotaAction
    ) -> CreateRegistryQuotaActionResult:
        quota_service = self._ensure_quota_service()
        await quota_service.create_quota(action.scope_id, action.quota)
        return CreateRegistryQuotaActionResult()

    async def read_registry_quota(
        self, action: ReadRegistryQuotaAction
    ) -> ReadRegistryQuotaActionResult:
        quota_service = self._ensure_quota_service()
        quota = await quota_service.read_quota(action.scope_id)
        return ReadRegistryQuotaActionResult(quota=quota)

    async def update_registry_quota(
        self, action: UpdateRegistryQuotaAction
    ) -> UpdateRegistryQuotaActionResult:
        quota_service = self._ensure_quota_service()
        await quota_service.update_quota(action.scope_id, action.quota)
        return UpdateRegistryQuotaActionResult()

    async def delete_registry_quota(
        self, action: DeleteRegistryQuotaAction
    ) -> DeleteRegistryQuotaActionResult:
        quota_service = self._ensure_quota_service()
        await quota_service.delete_quota(action.scope_id)
        return DeleteRegistryQuotaActionResult()



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager

import aiohttp_cors

from ai.backend.manager.api.rest.types import CORSOptions

from .base import SystemDependency


class CORSOptionsDependency(SystemDependency[CORSOptions]):
    """Provides CORS options configuration."""

    @property
    def stage_name(self) -> str:
        return "cors-options"

    @asynccontextmanager
    async def provide(self, setup_input: object) -> AsyncIterator[CORSOptions]:
        yield {
            "*": aiohttp_cors.ResourceOptions(  # type: ignore[no-untyped-call]
                allow_credentials=False, expose_headers="*", allow_headers="*"
            ),
        }



from dataclasses import dataclass
from typing import Self

from ai.backend.manager.repositories.session.repository import SessionRepository
from ai.backend.manager.repositories.types import RepositoryArgs


@dataclass
class SessionRepositories:
    repository: SessionRepository

    @classmethod
    def create(cls, args: RepositoryArgs) -> Self:
        repository = SessionRepository(args.db)

        return cls(
            repository=repository,
        )



from __future__ import annotations

from typing import TYPE_CHECKING
from uuid import UUID

from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience import (
    MetricArgs,
    MetricPolicy,
    Resilience,
    RetryArgs,
    RetryPolicy,
)
from ai.backend.common.resilience.policies.retry import BackoffStrategy
from ai.backend.manager.data.prometheus_query_preset import (
    PrometheusQueryPresetData,
    PrometheusQueryPresetListResult,
)
from ai.backend.manager.repositories.base import BatchQuerier, Creator
from ai.backend.manager.repositories.base.updater import Updater

from .db_source import PrometheusQueryPresetDBSource

if TYPE_CHECKING:
    from ai.backend.manager.models.prometheus_query_preset import (
        PrometheusQueryPresetRow,
    )
    from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

__all__ = ("PrometheusQueryPresetRepository",)


prometheus_query_preset_repository_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(
                domain=DomainType.REPOSITORY,
                layer=LayerType.PROMETHEUS_QUERY_PRESET_REPOSITORY,
            )
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=10,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
            )
        ),
    ]
)


class PrometheusQueryPresetRepository:
    """Repository for prometheus query preset data access."""

    _db_source: PrometheusQueryPresetDBSource

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db_source = PrometheusQueryPresetDBSource(db)

    @prometheus_query_preset_repository_resilience.apply()
    async def create(
        self,
        creator: Creator[PrometheusQueryPresetRow],
    ) -> PrometheusQueryPresetData:
        """Creates a new prometheus query preset."""
        return await self._db_source.create(creator)

    @prometheus_query_preset_repository_resilience.apply()
    async def update(
        self,
        updater: Updater[PrometheusQueryPresetRow],
    ) -> PrometheusQueryPresetData:
        """Updates an existing prometheus query preset."""
        return await self._db_source.update(updater=updater)

    @prometheus_query_preset_repository_resilience.apply()
    async def delete(self, preset_id: UUID) -> bool:
        """Deletes a prometheus query preset."""
        return await self._db_source.delete(preset_id)

    @prometheus_query_preset_repository_resilience.apply()
    async def get_by_id(self, preset_id: UUID) -> PrometheusQueryPresetData:
        """Retrieves a prometheus query preset by ID."""
        return await self._db_source.get_by_id(preset_id)

    @prometheus_query_preset_repository_resilience.apply()
    async def search(
        self,
        querier: BatchQuerier,
    ) -> PrometheusQueryPresetListResult:
        """Searches prometheus query presets with total count."""
        return await self._db_source.search(querier=querier)



"""Public health check handler using constructor dependency injection."""

from __future__ import annotations

from typing import TYPE_CHECKING

from aiohttp import web

from ai.backend.common.dto.internal.health import HealthResponse, HealthStatus
from ai.backend.manager import __version__

if TYPE_CHECKING:
    from ai.backend.common.health_checker.probe import HealthProbe


class HealthHandler:
    _health_probe: HealthProbe

    def __init__(self, *, health_probe: HealthProbe) -> None:
        self._health_probe = health_probe

    async def hello(self, request: web.Request) -> web.Response:
        """Health check endpoint with dependency connectivity status."""
        request["do_not_print_access_log"] = True
        connectivity = await self._health_probe.get_connectivity_status()
        response = HealthResponse(
            status=HealthStatus.OK if connectivity.overall_healthy else HealthStatus.DEGRADED,
            version=__version__,
            component="manager",
            connectivity=connectivity,
        )
        return web.json_response(response.model_dump(mode="json"))



"""Service (model serving) handler class using constructor dependency injection.

All handlers use the new ApiHandler pattern: typed parameters
(``BodyParam``, ``QueryParam``, ``PathParam``, ``UserContext``, ``RequestCtx``)
are automatically extracted by ``_wrap_api_handler`` and responses are
returned as ``APIResponse`` objects.
"""

from __future__ import annotations

import logging
import uuid
from http import HTTPStatus
from typing import TYPE_CHECKING, Any, Final

import yarl
from aiohttp import web
from pydantic import HttpUrl

from ai.backend.common.api_handlers import (
    APIResponse,
    BodyParam,
    PathParam,
    QueryParam,
)
from ai.backend.common.dto.manager.model_serving.request import (
    ListServeRequestModel,
    NewServiceRequestModel,
    ScaleRequestModel,
    SearchServicesRequestModel,
    ServiceIdPathParam,
    ServiceRouteIdPathParam,
    TokenRequestModel,
    UpdateRouteRequestModel,
)
from ai.backend.common.dto.manager.model_serving.response import (
    CompactServeInfoModel,
    ErrorInfoModel,
    ErrorListResponseModel,
    PaginationInfoModel,
    RouteInfoModel,
    RuntimeInfo,
    RuntimeInfoModel,
    ScaleResponseModel,
    SearchServicesResponseModel,
    ServeInfoModel,
    ServiceSearchItemModel,
    SuccessResponseModel,
    TokenResponseModel,
    TryStartResponseModel,
)
from ai.backend.common.types import (
    MODEL_SERVICE_RUNTIME_PROFILES,
    AccessKey,
    ClusterMode,
    MountPermission,
    MountTypes,
    RuntimeVariant,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.deployment.creator import DeploymentCreationDraft
from ai.backend.manager.data.deployment.types import (
    DeploymentInfo,
    DeploymentMetadata,
    DeploymentNetworkSpec,
    ExecutionSpec,
    ImageIdentifierDraft,
    ModelRevisionSpecDraft,
    MountMetadata,
    ReplicaSpec,
    ResourceSpecDraft,
)
from ai.backend.manager.data.model_serving.types import (
    ModelServicePrepareCtx,
    MountOption,
    ServiceConfig,
    ServiceInfo,
)
from ai.backend.manager.dto.context import RequestCtx, UserContext
from ai.backend.manager.services.auth.actions.resolve_access_key_scope import (
    ResolveAccessKeyScopeAction,
)
from ai.backend.manager.services.deployment.actions.create_legacy_deployment import (
    CreateLegacyDeploymentAction,
    CreateLegacyDeploymentActionResult,
)
from ai.backend.manager.services.deployment.actions.destroy_deployment import (
    DestroyDeploymentAction,
    DestroyDeploymentActionResult,
)
from ai.backend.manager.services.model_serving.actions.clear_error import ClearErrorAction
from ai.backend.manager.services.model_serving.actions.delete_route import DeleteRouteAction
from ai.backend.manager.services.model_serving.actions.dry_run_model_service import (
    DryRunModelServiceAction,
)
from ai.backend.manager.services.model_serving.actions.force_sync import ForceSyncAction
from ai.backend.manager.services.model_serving.actions.generate_token import GenerateTokenAction
from ai.backend.manager.services.model_serving.actions.get_model_service_info import (
    GetModelServiceInfoAction,
    GetModelServiceInfoActionResult,
)
from ai.backend.manager.services.model_serving.actions.list_errors import ListErrorsAction
from ai.backend.manager.services.model_serving.actions.list_model_service import (
    ListModelServiceAction,
    ListModelServiceActionResult,
)
from ai.backend.manager.services.model_serving.actions.scale_service_replicas import (
    ScaleServiceReplicasAction,
)
from ai.backend.manager.services.model_serving.actions.search_services import (
    SearchServicesAction,
    SearchServicesActionResult,
)
from ai.backend.manager.services.model_serving.actions.update_route import UpdateRouteAction
from ai.backend.manager.services.model_serving.actions.validate_model_service import (
    ValidateModelServiceAction,
    ValidateModelServiceActionResult,
)
from ai.backend.manager.services.model_serving.adapter import ServiceSearchAdapter
from ai.backend.manager.services.model_serving.processors.auto_scaling import (
    ModelServingAutoScalingProcessors,
)
from ai.backend.manager.services.model_serving.processors.model_serving import (
    ModelServingProcessors,
)

if TYPE_CHECKING:
    from ai.backend.manager.services.auth.processors import AuthProcessors
    from ai.backend.manager.services.deployment.processors import DeploymentProcessors

log: Final = BraceStyleAdapter(logging.getLogger(__spec__.name))


def _serve_info_from_dto(dto: ServiceInfo) -> ServeInfoModel:
    return ServeInfoModel(
        endpoint_id=dto.endpoint_id,
        model_id=dto.model_id,
        extra_mounts=dto.extra_mounts,
        name=dto.name,
        model_definition_path=dto.model_definition_path,
        replicas=dto.replicas,
        desired_session_count=dto.replicas,
        active_routes=[
            RouteInfoModel(
                route_id=route.route_id,
                session_id=route.session_id,
                traffic_ratio=route.traffic_ratio,
            )
            for route in dto.active_routes
        ],
        service_endpoint=dto.service_endpoint,
        is_public=dto.is_public,
        runtime_variant=dto.runtime_variant,
    )


def _serve_info_from_deployment_info(deployment_info: DeploymentInfo) -> ServeInfoModel:
    """Convert DeploymentInfo to ServeInfoModel."""
    model_revision = deployment_info.model_revisions[0] if deployment_info.model_revisions else None

    return ServeInfoModel(
        endpoint_id=deployment_info.id,
        model_id=model_revision.mounts.model_vfolder_id if model_revision else uuid.UUID(int=0),
        extra_mounts=[m.vfid.folder_id for m in model_revision.mounts.extra_mounts]
        if model_revision
        else [],
        name=deployment_info.metadata.name,
        model_definition_path=model_revision.mounts.model_definition_path
        if model_revision
        else None,
        replicas=deployment_info.replica_spec.replica_count,
        desired_session_count=deployment_info.replica_spec.replica_count,
        active_routes=[],
        service_endpoint=HttpUrl(deployment_info.network.url)
        if deployment_info.network.url
        else None,
        is_public=deployment_info.network.open_to_public,
        runtime_variant=model_revision.execution.runtime_variant
        if model_revision
        else RuntimeVariant.CUSTOM,
    )


class ServiceHandler:
    """Service (model serving) API handler with constructor-injected dependencies."""

    def __init__(
        self,
        *,
        auth: AuthProcessors,
        deployment: DeploymentProcessors,
        model_serving: ModelServingProcessors,
        model_serving_auto_scaling: ModelServingAutoScalingProcessors,
    ) -> None:
        self._auth = auth
        self._deployment = deployment
        self._model_serving = model_serving
        self._model_serving_auto_scaling = model_serving_auto_scaling

    # ------------------------------------------------------------------
    # list_serve (GET /services)
    # ------------------------------------------------------------------

    async def list_serve(
        self, query: QueryParam[ListServeRequestModel], ctx: UserContext
    ) -> APIResponse | web.StreamResponse:
        params = query.parsed
        log.info("SERVE.LIST (email:{}, ak:{})", ctx.user_email, ctx.access_key)

        action = ListModelServiceAction(session_owener_id=ctx.user_uuid, name=params.name)
        result: ListModelServiceActionResult = (
            await self._model_serving.list_model_service.wait_for_complete(action)
        )

        items = [
            CompactServeInfoModel(
                id=info.id,
                name=info.name,
                replicas=info.replicas,
                desired_session_count=info.replicas,
                active_route_count=info.active_route_count,
                service_endpoint=info.service_endpoint,
                is_public=info.is_public,
            )
            for info in result.data
        ]
        return web.json_response(
            [item.model_dump(mode="json") for item in items],
            status=HTTPStatus.OK,
        )

    # ------------------------------------------------------------------
    # search_services (POST /services/_/search)
    # ------------------------------------------------------------------

    async def search_services(
        self, body: BodyParam[SearchServicesRequestModel], ctx: UserContext
    ) -> APIResponse:
        params = body.parsed
        log.info("SERVE.SEARCH (email:{}, ak:{})", ctx.user_email, ctx.access_key)

        adapter = ServiceSearchAdapter()
        conditions = adapter.convert_filter(params.filter) if params.filter else []

        action = SearchServicesAction(
            session_owner_id=ctx.user_uuid,
            conditions=conditions,
            offset=params.offset,
            limit=params.limit,
        )
        result: SearchServicesActionResult = (
            await self._model_serving.search_services.wait_for_complete(action)
        )

        resp = SearchServicesResponseModel(
            items=[
                ServiceSearchItemModel(
                    id=item.id,
                    name=item.name,
                    desired_session_count=item.replicas,
                    replicas=item.replicas,
                    active_route_count=item.active_route_count,
                    service_endpoint=item.service_endpoint,
                    resource_slots=dict(item.resource_slots),
                    resource_group=item.resource_group,
                    open_to_public=item.open_to_public,
                )
                for item in result.items
            ],
            pagination=PaginationInfoModel(
                total=result.total_count,
                offset=result.offset,
                limit=result.limit,
            ),
        )
        return APIResponse.build(HTTPStatus.OK, resp)

    # ------------------------------------------------------------------
    # get_info (GET /services/{service_id})
    # ------------------------------------------------------------------

    async def get_info(self, path: PathParam[ServiceIdPathParam], ctx: UserContext) -> APIResponse:
        params = path.parsed
        log.info(
            "SERVE.GET_INFO (email:{}, ak:{}, s:{})",
            ctx.user_email,
            ctx.access_key,
            params.service_id,
        )

        action = GetModelServiceInfoAction(service_id=params.service_id)
        result: GetModelServiceInfoActionResult = (
            await self._model_serving.get_model_service_info.wait_for_complete(action)
        )

        resp = _serve_info_from_dto(result.data)
        return APIResponse.build(HTTPStatus.OK, resp)

    # ------------------------------------------------------------------
    # create (POST /services)
    # ------------------------------------------------------------------

    async def create(self, body: BodyParam[NewServiceRequestModel], req: RequestCtx) -> APIResponse:
        params = body.parsed
        request = req.request
        validation_result = await self._run_validation(request, params)

        deployment_action = CreateLegacyDeploymentAction(
            draft=DeploymentCreationDraft(
                metadata=DeploymentMetadata(
                    name=params.service_name,
                    domain=params.domain_name,
                    project=validation_result.group_id,
                    resource_group=validation_result.scaling_group,
                    created_user=request["user"]["uuid"],
                    session_owner=validation_result.owner_uuid,
                    created_at=None,
                    revision_history_limit=10,
                    tag=params.tag,
                ),
                replica_spec=ReplicaSpec(
                    replica_count=params.replicas,
                ),
                draft_model_revision=self._to_model_revision(params, validation_result),
                network=DeploymentNetworkSpec(
                    open_to_public=params.open_to_public,
                ),
            )
        )
        deployment_result: CreateLegacyDeploymentActionResult = (
            await self._deployment.create_legacy_deployment.wait_for_complete(deployment_action)
        )
        resp = _serve_info_from_deployment_info(deployment_result.data)
        return APIResponse.build(HTTPStatus.CREATED, resp)

    # ------------------------------------------------------------------
    # try_start (POST /services/_/try)
    # ------------------------------------------------------------------

    async def try_start(
        self, body: BodyParam[NewServiceRequestModel], req: RequestCtx
    ) -> APIResponse:
        params = body.parsed
        request = req.request
        validation_result = await self._run_validation(request, params)

        action = self._to_start_action(params, validation_result, request)
        result = await self._model_serving.dry_run_model_service.wait_for_complete(action)

        resp = TryStartResponseModel(task_id=str(result.task_id))
        return APIResponse.build(HTTPStatus.OK, resp)

    # ------------------------------------------------------------------
    # list_supported_runtimes (GET /services/_/runtimes)
    # ------------------------------------------------------------------

    async def list_supported_runtimes(self, ctx: UserContext) -> APIResponse:
        resp = RuntimeInfoModel(
            runtimes=[
                RuntimeInfo(
                    name=v.value,
                    human_readable_name=MODEL_SERVICE_RUNTIME_PROFILES[v].name,
                )
                for v in RuntimeVariant
            ]
        )
        return APIResponse.build(HTTPStatus.OK, resp)

    # ------------------------------------------------------------------
    # delete (DELETE /services/{service_id})
    # ------------------------------------------------------------------

    async def delete(self, path: PathParam[ServiceIdPathParam], ctx: UserContext) -> APIResponse:
        params = path.parsed
        log.info(
            "SERVE.DELETE (email:{}, ak:{}, s:{})",
            ctx.user_email,
            ctx.access_key,
            params.service_id,
        )

        deployment_action = DestroyDeploymentAction(endpoint_id=params.service_id)
        deployment_result: DestroyDeploymentActionResult = (
            await self._deployment.destroy_deployment.wait_for_complete(deployment_action)
        )
        resp = SuccessResponseModel(success=deployment_result.success)
        return APIResponse.build(HTTPStatus.OK, resp)

    # ------------------------------------------------------------------
    # sync (POST /services/{service_id}/sync)
    # ------------------------------------------------------------------

    async def sync(self, path: PathParam[ServiceIdPathParam], ctx: UserContext) -> APIResponse:
        params = path.parsed
        log.info(
            "SERVE.SYNC (email:{}, ak:{}, s:{})",
            ctx.user_email,
            ctx.access_key,
            params.service_id,
        )

        action = ForceSyncAction(service_id=params.service_id)
        result = await self._model_serving.force_sync.wait_for_complete(action)

        resp = SuccessResponseModel(success=result.success)
        return APIResponse.build(HTTPStatus.OK, resp)

    # ------------------------------------------------------------------
    # scale (POST /services/{service_id}/scale)
    # ------------------------------------------------------------------

    async def scale(
        self,
        path: PathParam[ServiceIdPathParam],
        body: BodyParam[ScaleRequestModel],
        req: RequestCtx,
    ) -> APIResponse:
        path_params = path.parsed
        params = body.parsed
        request = req.request
        log.info(
            "SERVE.SCALE (email:{}, ak:{}, s:{})",
            request["user"]["email"],
            request["keypair"]["access_key"],
            path_params.service_id,
        )

        action = ScaleServiceReplicasAction(
            max_session_count_per_model_session=request["user"]["resource_policy"][
                "max_session_count_per_model_session"
            ],
            service_id=path_params.service_id,
            to=params.to,
        )

        result = await self._model_serving_auto_scaling.scale_service_replicas.wait_for_complete(
            action
        )

        resp = ScaleResponseModel(
            current_route_count=result.current_route_count,
            target_count=result.target_count,
        )
        return APIResponse.build(HTTPStatus.OK, resp)

    # ------------------------------------------------------------------
    # update_route (PUT /services/{service_id}/routings/{route_id})
    # ------------------------------------------------------------------

    async def update_route(
        self,
        path: PathParam[ServiceRouteIdPathParam],
        body: BodyParam[UpdateRouteRequestModel],
        ctx: UserContext,
    ) -> APIResponse:
        path_params = path.parsed
        params = body.parsed
        log.info(
            "SERVE.UPDATE_ROUTE (email:{}, ak:{}, s:{}, r:{})",
            ctx.user_email,
            ctx.access_key,
            path_params.service_id,
            path_params.route_id,
        )

        action = UpdateRouteAction(
            service_id=path_params.service_id,
            route_id=path_params.route_id,
            traffic_ratio=params.traffic_ratio,
        )

        result = await self._model_serving.update_route.wait_for_complete(action)

        resp = SuccessResponseModel(success=result.success)
        return APIResponse.build(HTTPStatus.OK, resp)

    # ------------------------------------------------------------------
    # delete_route (DELETE /services/{service_id}/routings/{route_id})
    # ------------------------------------------------------------------

    async def delete_route(
        self, path: PathParam[ServiceRouteIdPathParam], ctx: UserContext
    ) -> APIResponse:
        path_params = path.parsed
        log.info(
            "SERVE.DELETE_ROUTE (email:{}, ak:{}, s:{})",
            ctx.user_email,
            ctx.access_key,
            path_params.service_id,
        )

        action = DeleteRouteAction(
            service_id=path_params.service_id,
            route_id=path_params.route_id,
        )

        result = await self._model_serving.delete_route.wait_for_complete(action)

        resp = SuccessResponseModel(success=result.success)
        return APIResponse.build(HTTPStatus.OK, resp)

    # ------------------------------------------------------------------
    # generate_token (POST /services/{service_id}/token)
    # ------------------------------------------------------------------

    async def generate_token(
        self,
        path: PathParam[ServiceIdPathParam],
        body: BodyParam[TokenRequestModel],
        ctx: UserContext,
    ) -> APIResponse:
        path_params = path.parsed
        params = body.parsed
        log.info(
            "SERVE.GENERATE_TOKEN (email:{}, ak:{}, s:{})",
            ctx.user_email,
            ctx.access_key,
            path_params.service_id,
        )

        action = GenerateTokenAction(
            service_id=path_params.service_id,
            duration=params.duration,
            valid_until=params.valid_until,
            expires_at=params.expires_at,
        )

        result = await self._model_serving.generate_token.wait_for_complete(action)

        resp = TokenResponseModel(token=result.data.token)
        return APIResponse.build(HTTPStatus.OK, resp)

    # ------------------------------------------------------------------
    # list_errors (GET /services/{service_id}/errors)
    # ------------------------------------------------------------------

    async def list_errors(
        self, path: PathParam[ServiceIdPathParam], ctx: UserContext
    ) -> APIResponse:
        path_params = path.parsed
        log.info(
            "SERVE.LIST_ERRORS (email:{}, ak:{}, s:{})",
            ctx.user_email,
            ctx.access_key,
            path_params.service_id,
        )

        action = ListErrorsAction(service_id=path_params.service_id)
        result = await self._model_serving.list_errors.wait_for_complete(action)

        resp = ErrorListResponseModel(
            errors=[
                ErrorInfoModel(session_id=info.session_id, error=info.error)
                for info in result.error_info
            ],
            retries=result.retries,
        )
        return APIResponse.build(HTTPStatus.OK, resp)

    # ------------------------------------------------------------------
    # clear_error (POST /services/{service_id}/errors/clear)
    # ------------------------------------------------------------------

    async def clear_error(
        self, path: PathParam[ServiceIdPathParam], ctx: UserContext
    ) -> APIResponse:
        path_params = path.parsed
        log.info(
            "SERVE.CLEAR_ERROR (email:{}, ak:{}, s:{})",
            ctx.user_email,
            ctx.access_key,
            path_params.service_id,
        )

        action = ClearErrorAction(service_id=path_params.service_id)
        await self._model_serving.clear_error.wait_for_complete(action)

        return APIResponse.no_content(HTTPStatus.NO_CONTENT)

    # ------------------------------------------------------------------
    # Private helpers
    # ------------------------------------------------------------------

    async def _run_validation(
        self,
        request: Any,
        params: NewServiceRequestModel,
    ) -> ValidateModelServiceActionResult:
        scope = await self._auth.resolve_access_key_scope.wait_for_complete(
            ResolveAccessKeyScopeAction(
                requester_access_key=request["keypair"]["access_key"],
                requester_role=request["user"]["role"],
                requester_domain=request["user"]["domain_name"],
                owner_access_key=params.owner_access_key,
            )
        )
        requester_access_key, owner_access_key = (
            scope.requester_access_key,
            scope.owner_access_key,
        )
        action = ValidateModelServiceAction(
            requester_access_key=requester_access_key,
            owner_access_key=owner_access_key,
            requester_uuid=request["user"]["uuid"],
            requester_role=request["user"]["role"],
            requester_domain=request["user"]["domain_name"],
            keypair_resource_policy=request["user"]["resource_policy"],
            domain_name=params.domain_name,
            group_name=params.group_name,
            config=ServiceConfig(
                model=params.config.model,
                model_definition_path=params.config.model_definition_path,
                model_version=params.config.model_version,
                model_mount_destination=params.config.model_mount_destination,
                extra_mounts={
                    k: MountOption(
                        mount_destination=v.get("mount_destination"),
                        type=MountTypes(v["type"]) if v.get("type") else MountTypes.BIND,
                        permission=MountPermission(v["permission"])
                        if v.get("permission")
                        else None,
                    )
                    for k, v in params.config.extra_mounts.items()
                },
                environ=params.config.environ,
                scaling_group=params.config.scaling_group,
                resources=params.config.resources,
                resource_opts=params.config.resource_opts,
            ),
            replicas=params.replicas,
            runtime_variant=params.runtime_variant,
            max_session_count_per_model_session=request["user"]["resource_policy"][
                "max_session_count_per_model_session"
            ],
            owner_access_key_override=AccessKey(params.owner_access_key)
            if params.owner_access_key
            else None,
        )
        return await self._model_serving.validate_model_service.wait_for_complete(action)

    @staticmethod
    def _to_model_revision(
        params: NewServiceRequestModel,
        validation_result: ValidateModelServiceActionResult,
    ) -> ModelRevisionSpecDraft:
        return ModelRevisionSpecDraft(
            image_identifier=ImageIdentifierDraft(
                canonical=params.image,
                architecture=params.architecture,
            ),
            resource_spec=ResourceSpecDraft(
                cluster_mode=ClusterMode(params.cluster_mode),
                cluster_size=params.cluster_size,
                resource_slots=params.config.resources,
                resource_opts=params.config.resource_opts,
            ),
            mounts=MountMetadata(
                model_vfolder_id=validation_result.model_id,
                model_definition_path=validation_result.model_definition_path,
                model_mount_destination=params.config.model_mount_destination,
                extra_mounts=list(validation_result.extra_mounts),
            ),
            execution=ExecutionSpec(
                startup_command=params.startup_command,
                bootstrap_script=params.bootstrap_script,
                environ=params.config.environ,
                runtime_variant=params.runtime_variant,
                callback_url=yarl.URL(str(params.callback_url)) if params.callback_url else None,
            ),
        )

    @staticmethod
    def _to_start_action(
        params: NewServiceRequestModel,
        validation_result: ValidateModelServiceActionResult,
        request: Any,
    ) -> DryRunModelServiceAction:
        return DryRunModelServiceAction(
            service_name=params.service_name,
            replicas=params.replicas,
            image=params.image,
            runtime_variant=params.runtime_variant,
            architecture=params.architecture,
            group_name=params.group_name,
            domain_name=params.domain_name,
            cluster_size=params.cluster_size,
            cluster_mode=ClusterMode(params.cluster_mode),
            tag=params.tag,
            startup_command=params.startup_command,
            bootstrap_script=params.bootstrap_script,
            callback_url=params.callback_url,
            owner_access_key=params.owner_access_key,
            open_to_public=params.open_to_public,
            config=ServiceConfig(
                model=params.config.model,
                model_definition_path=params.config.model_definition_path,
                model_version=params.config.model_version,
                model_mount_destination=params.config.model_mount_destination,
                extra_mounts={
                    k: MountOption(
                        mount_destination=v.get("mount_destination"),
                        type=MountTypes(v["type"]) if v.get("type") else MountTypes.BIND,
                        permission=MountPermission(v["permission"])
                        if v.get("permission")
                        else None,
                    )
                    for k, v in params.config.extra_mounts.items()
                },
                environ=params.config.environ,
                scaling_group=params.config.scaling_group,
                resources=params.config.resources,
                resource_opts=params.config.resource_opts,
            ),
            request_user_id=request["user"]["uuid"],
            sudo_session_enabled=request["user"]["sudo_session_enabled"],
            model_service_prepare_ctx=ModelServicePrepareCtx(
                model_id=validation_result.model_id,
                model_definition_path=validation_result.model_definition_path,
                requester_access_key=validation_result.requester_access_key,
                owner_access_key=validation_result.owner_access_key,
                owner_uuid=validation_result.owner_uuid,
                owner_role=validation_result.owner_role,
                group_id=validation_result.group_id,
                resource_policy=validation_result.resource_policy,
                scaling_group=validation_result.scaling_group,
                extra_mounts=validation_result.extra_mounts,
            ),
        )



from __future__ import annotations

from typing import override

from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.processor import ActionProcessor
from ai.backend.manager.actions.types import AbstractProcessorPackage, ActionSpec
from ai.backend.manager.actions.validators import ActionValidators
from ai.backend.manager.services.service_catalog.actions.search import (
    SearchServiceCatalogsAction,
    SearchServiceCatalogsActionResult,
)
from ai.backend.manager.services.service_catalog.service import ServiceCatalogService


class ServiceCatalogProcessors(AbstractProcessorPackage):
    """Processor package for service catalog operations."""

    search_service_catalogs: ActionProcessor[
        SearchServiceCatalogsAction, SearchServiceCatalogsActionResult
    ]

    def __init__(
        self,
        service: ServiceCatalogService,
        action_monitors: list[ActionMonitor],
        validators: ActionValidators,
    ) -> None:
        self.search_service_catalogs = ActionProcessor(
            service.search_service_catalogs, action_monitors
        )

    @override
    def supported_actions(self) -> list[ActionSpec]:
        return [
            SearchServiceCatalogsAction.spec(),
        ]



from .handler import PrometheusQueryPresetHandler
from .registry import register_prometheus_query_preset_routes

__all__ = [
    "PrometheusQueryPresetHandler",
    "register_prometheus_query_preset_routes",
]



from __future__ import annotations

from collections.abc import Collection
from dataclasses import dataclass
from typing import Any, cast
from uuid import UUID

import sqlalchemy as sa
from sqlalchemy.engine import CursorResult
from sqlalchemy.ext.asyncio import AsyncSession as SASession

from ai.backend.common.data.permission.types import OperationType, ScopeType
from ai.backend.manager.data.permission.id import ObjectId
from ai.backend.manager.models.rbac_models.permission.permission import PermissionRow

# =============================================================================
# Data Classes
# =============================================================================


@dataclass
class RBACRevoker:
    """
    Data class for revoking permissions from specific role(s).

    Uses entity-as-scope pattern: permissions are stored with scope_type/scope_id
    referring to the entity being revoked.

    Note: Only entity-level revoking is supported. Field-level revoking is not supported.

    Attributes:
        entity_id: The entity to revoke access from.
        entity_scope_type: The scope type for entity-as-scope in permissions table.
        target_role_ids: The role ID(s) to revoke permissions from.
        operations: Operations to revoke. If None, revokes all operations for the entity.
    """

    entity_id: ObjectId
    entity_scope_type: ScopeType
    target_role_ids: list[UUID]
    operations: list[OperationType] | None = None


# =============================================================================
# Deletion Helpers
# =============================================================================


async def _delete_permissions(
    db_sess: SASession,
    role_ids: Collection[UUID],
    entity_id: ObjectId,
    scope_type: ScopeType,
    operations: list[OperationType] | None,
) -> int:
    """Delete permissions for the given entity-as-scope and roles."""
    if not role_ids:
        return 0

    conditions = [
        PermissionRow.role_id.in_(role_ids),
        PermissionRow.scope_type == scope_type,
        PermissionRow.scope_id == entity_id.entity_id,
        PermissionRow.entity_type == entity_id.entity_type,
    ]

    if operations is not None:
        conditions.append(PermissionRow.operation.in_(operations))

    result = await db_sess.execute(sa.delete(PermissionRow).where(sa.and_(*conditions)))
    return cast(CursorResult[Any], result).rowcount or 0


# =============================================================================
# Public API
# =============================================================================


async def execute_rbac_revoker(
    db_sess: SASession,
    revoker: RBACRevoker,
) -> None:
    """
    Revoke permissions from specified roles using entity-as-scope pattern.

    This is used when unsharing an entity from specific roles.
    For example, when user A revokes user B's access to a VFolder:
    - User B's role loses permissions for that VFolder.

    Args:
        db_sess: Async SQLAlchemy session (must be writable).
        revoker: Revoker instance containing entity_id, entity_scope_type,
                 target_role_ids, and optional operations.
    """
    if not revoker.target_role_ids:
        return

    await _delete_permissions(
        db_sess,
        revoker.target_role_ids,
        revoker.entity_id,
        revoker.entity_scope_type,
        revoker.operations,
    )

    await db_sess.flush()



from __future__ import annotations

import uuid
from typing import TYPE_CHECKING

from ai.backend.common.exception import BackendAIError
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience
from ai.backend.manager.data.error_log.types import ErrorLogData, ErrorLogListResult
from ai.backend.manager.models.error_logs import ErrorLogRow
from ai.backend.manager.repositories.base import BatchQuerier, Creator

from .db_source import ErrorLogDBSource

if TYPE_CHECKING:
    from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

__all__ = ("ErrorLogRepository",)

error_log_repository_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(domain=DomainType.REPOSITORY, layer=LayerType.ERROR_LOG_REPOSITORY)
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=10,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class ErrorLogRepository:
    _db_source: ErrorLogDBSource

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db_source = ErrorLogDBSource(db)

    @error_log_repository_resilience.apply()
    async def create(self, creator: Creator[ErrorLogRow]) -> ErrorLogData:
        return await self._db_source.create(creator)

    @error_log_repository_resilience.apply()
    async def search(
        self,
        querier: BatchQuerier,
    ) -> ErrorLogListResult:
        """Search error logs with querier pattern."""
        return await self._db_source.search(querier=querier)

    @error_log_repository_resilience.apply()
    async def list_logs(
        self,
        *,
        user_uuid: uuid.UUID,
        user_domain: str,
        is_superadmin: bool,
        is_admin: bool,
        page_no: int,
        page_size: int,
        mark_read: bool,
    ) -> tuple[list[ErrorLogData], int]:
        """List error logs with role-based visibility filtering."""
        return await self._db_source.list_logs(
            user_uuid=user_uuid,
            user_domain=user_domain,
            is_superadmin=is_superadmin,
            is_admin=is_admin,
            page_no=page_no,
            page_size=page_size,
            mark_read=mark_read,
        )

    @error_log_repository_resilience.apply()
    async def mark_cleared(
        self,
        *,
        log_id: uuid.UUID,
        user_uuid: uuid.UUID,
        user_domain: str,
        is_superadmin: bool,
        is_admin: bool,
    ) -> int:
        """Mark an error log as cleared. Returns number of rows updated."""
        return await self._db_source.mark_cleared(
            log_id=log_id,
            user_uuid=user_uuid,
            user_domain=user_domain,
            is_superadmin=is_superadmin,
            is_admin=is_admin,
        )



"""Prometheus query preset GQL fetcher functions."""

from .preset import (
    fetch_prometheus_query_preset,
    fetch_prometheus_query_preset_result,
    fetch_prometheus_query_presets,
)

__all__ = [
    "fetch_prometheus_query_preset",
    "fetch_prometheus_query_presets",
    "fetch_prometheus_query_preset_result",
]



import logging

from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.clients.storage_proxy.session_manager import StorageSessionManager
from ai.backend.manager.errors.common import GenericBadRequest
from ai.backend.manager.repositories.vfs_storage.repository import VFSStorageRepository
from ai.backend.manager.services.vfs_storage.actions.create import (
    CreateVFSStorageAction,
    CreateVFSStorageActionResult,
)
from ai.backend.manager.services.vfs_storage.actions.delete import (
    DeleteVFSStorageAction,
    DeleteVFSStorageActionResult,
)
from ai.backend.manager.services.vfs_storage.actions.get import (
    GetVFSStorageAction,
    GetVFSStorageActionResult,
)
from ai.backend.manager.services.vfs_storage.actions.get_quota_scope import (
    GetQuotaScopeAction,
    GetQuotaScopeActionResult,
)
from ai.backend.manager.services.vfs_storage.actions.list import (
    ListVFSStorageAction,
    ListVFSStorageActionResult,
)
from ai.backend.manager.services.vfs_storage.actions.search import (
    SearchVFSStoragesAction,
    SearchVFSStoragesActionResult,
)
from ai.backend.manager.services.vfs_storage.actions.search_quota_scopes import (
    QuotaScopeInfo,
    SearchQuotaScopesAction,
    SearchQuotaScopesActionResult,
)
from ai.backend.manager.services.vfs_storage.actions.set_quota_scope import (
    SetQuotaScopeAction,
    SetQuotaScopeActionResult,
)
from ai.backend.manager.services.vfs_storage.actions.unset_quota_scope import (
    UnsetQuotaScopeAction,
    UnsetQuotaScopeActionResult,
)
from ai.backend.manager.services.vfs_storage.actions.update import (
    UpdateVFSStorageAction,
    UpdateVFSStorageActionResult,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class VFSStorageService:
    _vfs_storage_repository: VFSStorageRepository
    _storage_manager: StorageSessionManager | None

    def __init__(
        self,
        vfs_storage_repository: VFSStorageRepository,
        storage_manager: StorageSessionManager | None = None,
    ) -> None:
        self._vfs_storage_repository = vfs_storage_repository
        self._storage_manager = storage_manager

    async def create(self, action: CreateVFSStorageAction) -> CreateVFSStorageActionResult:
        """
        Create a new VFS storage.
        """
        log.info("Creating VFS storage with data: {}", action.creator)
        storage_data = await self._vfs_storage_repository.create(action.creator)
        return CreateVFSStorageActionResult(result=storage_data)

    async def update(self, action: UpdateVFSStorageAction) -> UpdateVFSStorageActionResult:
        """
        Update an existing VFS storage.
        """
        log.info("Updating VFS storage with id: {}", action.updater.pk_value)
        storage_data = await self._vfs_storage_repository.update(action.updater)
        return UpdateVFSStorageActionResult(result=storage_data)

    async def delete(self, action: DeleteVFSStorageAction) -> DeleteVFSStorageActionResult:
        """
        Delete an existing VFS storage.
        """
        log.info("Deleting VFS storage with id: {}", action.storage_id)
        storage_data = await self._vfs_storage_repository.delete(action.storage_id)
        return DeleteVFSStorageActionResult(deleted_storage_id=storage_data)

    async def get(self, action: GetVFSStorageAction) -> GetVFSStorageActionResult:
        """
        Get an existing VFS storage by ID.
        """
        log.info("Getting VFS storage with id: {}", action.storage_id)
        if action.storage_id:
            storage_data = await self._vfs_storage_repository.get_by_id(action.storage_id)
        elif action.storage_name:
            storage_data = await self._vfs_storage_repository.get_by_name(action.storage_name)
        else:
            raise GenericBadRequest("Either storage_id or storage_name must be provided")

        return GetVFSStorageActionResult(result=storage_data)

    async def list(self, _action: ListVFSStorageAction) -> ListVFSStorageActionResult:
        """
        List all VFS storages.
        """
        log.info("Listing VFS storages")
        storage_data_list = await self._vfs_storage_repository.list_vfs_storages()
        return ListVFSStorageActionResult(data=storage_data_list)

    async def search(self, action: SearchVFSStoragesAction) -> SearchVFSStoragesActionResult:
        """
        Search VFS storages with pagination and filtering.
        """
        log.info("Searching VFS storages with querier: {}", action.querier)
        result = await self._vfs_storage_repository.search(action.querier)
        return SearchVFSStoragesActionResult(
            storages=result.items,
            total_count=result.total_count,
            has_next_page=result.has_next_page,
            has_previous_page=result.has_previous_page,
        )

    def _ensure_storage_manager(self) -> StorageSessionManager:
        if self._storage_manager is None:
            raise RuntimeError("Storage manager is not configured")
        return self._storage_manager

    async def get_quota_scope(self, action: GetQuotaScopeAction) -> GetQuotaScopeActionResult:
        storage_manager = self._ensure_storage_manager()
        proxy_name, volume_name = StorageSessionManager.get_proxy_and_volume(
            action.storage_host_name
        )
        manager_client = storage_manager.get_manager_facing_client(proxy_name)
        quota_config = await manager_client.get_quota_scope(volume_name, action.quota_scope_id)
        usage_bytes = quota_config.get("used_bytes")
        if usage_bytes is not None and usage_bytes < 0:
            usage_bytes = None
        return GetQuotaScopeActionResult(
            quota_scope_id=action.quota_scope_id,
            storage_host_name=action.storage_host_name,
            usage_bytes=usage_bytes,
            usage_count=None,
            hard_limit_bytes=quota_config.get("limit_bytes") or None,
        )

    async def search_quota_scopes(
        self, _action: SearchQuotaScopesAction
    ) -> SearchQuotaScopesActionResult:
        storage_manager = self._ensure_storage_manager()
        all_volumes = await storage_manager.get_all_volumes()
        quota_scopes: list[QuotaScopeInfo] = []
        for host, _volume_info in all_volumes:
            proxy_name, volume_name = StorageSessionManager.get_proxy_and_volume(host)
            manager_client = storage_manager.get_manager_facing_client(proxy_name)
            try:
                quota_config = await manager_client.get_quota_scope(volume_name, "")
                usage_bytes = quota_config.get("used_bytes")
                if usage_bytes is not None and usage_bytes < 0:
                    usage_bytes = None
                quota_scopes.append(
                    QuotaScopeInfo(
                        quota_scope_id="",
                        storage_host_name=host,
                        usage_bytes=usage_bytes,
                        usage_count=None,
                        hard_limit_bytes=quota_config.get("limit_bytes") or None,
                    )
                )
            except Exception:
                pass
        return SearchQuotaScopesActionResult(quota_scopes=quota_scopes)

    async def set_quota_scope(self, action: SetQuotaScopeAction) -> SetQuotaScopeActionResult:
        storage_manager = self._ensure_storage_manager()
        proxy_name, volume_name = StorageSessionManager.get_proxy_and_volume(
            action.storage_host_name
        )
        manager_client = storage_manager.get_manager_facing_client(proxy_name)
        await manager_client.update_quota_scope(
            volume_name, action.quota_scope_id, action.hard_limit_bytes
        )
        quota_config = await manager_client.get_quota_scope(volume_name, action.quota_scope_id)
        usage_bytes = quota_config.get("used_bytes")
        if usage_bytes is not None and usage_bytes < 0:
            usage_bytes = None
        return SetQuotaScopeActionResult(
            quota_scope_id=action.quota_scope_id,
            storage_host_name=action.storage_host_name,
            usage_bytes=usage_bytes,
            usage_count=None,
            hard_limit_bytes=quota_config.get("limit_bytes") or None,
        )

    async def unset_quota_scope(self, action: UnsetQuotaScopeAction) -> UnsetQuotaScopeActionResult:
        storage_manager = self._ensure_storage_manager()
        proxy_name, volume_name = StorageSessionManager.get_proxy_and_volume(
            action.storage_host_name
        )
        manager_client = storage_manager.get_manager_facing_client(proxy_name)
        await manager_client.delete_quota_scope_quota(volume_name, action.quota_scope_id)
        return UnsetQuotaScopeActionResult(
            quota_scope_id=action.quota_scope_id,
            storage_host_name=action.storage_host_name,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType

from .base import TemplateAction


@dataclass
class DeleteTaskTemplateAction(TemplateAction):
    """Action to soft-delete a task template."""

    template_id: str

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.DELETE

    @override
    def entity_id(self) -> str | None:
        return self.template_id


@dataclass
class DeleteTaskTemplateActionResult(BaseActionResult):
    """Result of deleting a task template."""

    @override
    def entity_id(self) -> str | None:
        return None



"""Resource preset API handler using constructor dependency injection.

All handlers use the new ApiHandler pattern: typed parameters
(``BodyParam``, ``QueryParam``, ``UserContext``) are automatically
extracted by ``_wrap_api_handler`` and responses are returned as
``APIResponse`` objects.
"""

from __future__ import annotations

import logging
from http import HTTPStatus
from typing import TYPE_CHECKING, Any, Final

from ai.backend.common.api_handlers import APIResponse, BodyParam, QueryParam
from ai.backend.common.dto.manager.resource.request import (
    CheckPresetsRequest,
    ListPresetsQuery,
    UsagePerMonthQuery,
    UsagePerPeriodQuery,
    WatcherAgentRequest,
    WatcherStatusQuery,
)
from ai.backend.common.dto.manager.resource.response import (
    CheckPresetsResponse,
    ContainerRegistriesResponse,
    EmptyResponse,
    ListPresetsResponse,
    RawListResponse,
    WatcherDataResponse,
)
from ai.backend.common.types import AccessKey, AgentId
from ai.backend.common.types import LegacyResourceSlotState as ResourceSlotState
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.dto.context import RequestCtx, UserContext
from ai.backend.manager.repositories.resource_slot.types import quantities_to_json
from ai.backend.manager.services.agent.actions.get_watcher_status import GetWatcherStatusAction
from ai.backend.manager.services.agent.actions.recalculate_usage import RecalculateUsageAction
from ai.backend.manager.services.agent.actions.watcher_agent_restart import (
    WatcherAgentRestartAction,
)
from ai.backend.manager.services.agent.actions.watcher_agent_start import WatcherAgentStartAction
from ai.backend.manager.services.agent.actions.watcher_agent_stop import WatcherAgentStopAction
from ai.backend.manager.services.container_registry.actions.get_container_registries import (
    GetContainerRegistriesAction,
)
from ai.backend.manager.services.group.actions.usage_per_month import UsagePerMonthAction
from ai.backend.manager.services.group.actions.usage_per_period import UsagePerPeriodAction
from ai.backend.manager.services.resource_preset.actions.check_presets import (
    CheckResourcePresetsAction,
)
from ai.backend.manager.services.resource_preset.actions.list_presets import (
    ListResourcePresetsAction,
)
from ai.backend.manager.services.user.actions.admin_month_stats import AdminMonthStatsAction
from ai.backend.manager.services.user.actions.user_month_stats import UserMonthStatsAction

if TYPE_CHECKING:
    from ai.backend.manager.services.agent.processors import AgentProcessors
    from ai.backend.manager.services.container_registry.processors import (
        ContainerRegistryProcessors,
    )
    from ai.backend.manager.services.group.processors import GroupProcessors
    from ai.backend.manager.services.resource_preset.processors import ResourcePresetProcessors
    from ai.backend.manager.services.user.processors import UserProcessors

log: Final = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ResourceHandler:
    """Resource preset API handler with constructor-injected dependencies."""

    def __init__(
        self,
        *,
        resource_preset: ResourcePresetProcessors,
        agent: AgentProcessors,
        group: GroupProcessors,
        user: UserProcessors,
        container_registry: ContainerRegistryProcessors,
    ) -> None:
        self._resource_preset = resource_preset
        self._agent = agent
        self._group = group
        self._user = user
        self._container_registry = container_registry

    # ------------------------------------------------------------------
    # list_presets (GET /resource/presets)
    # ------------------------------------------------------------------

    async def list_presets(
        self,
        query: QueryParam[ListPresetsQuery],
        ctx: UserContext,
    ) -> APIResponse:
        log.info("LIST_PRESETS (ak:{})", ctx.access_key)
        params = query.parsed
        result = await self._resource_preset.list_presets.wait_for_complete(
            ListResourcePresetsAction(
                access_key=ctx.access_key,
                scaling_group=params.scaling_group,
            )
        )
        return APIResponse.build(HTTPStatus.OK, ListPresetsResponse(presets=result.presets))

    # ------------------------------------------------------------------
    # check_presets (POST /resource/check-presets)
    # ------------------------------------------------------------------

    async def check_presets(
        self,
        body: BodyParam[CheckPresetsRequest],
        ctx: UserContext,
        req: RequestCtx,
    ) -> APIResponse:
        params = body.parsed
        resource_policy = req.request["keypair"]["resource_policy"]
        log.info(
            "CHECK_PRESETS (ak:{}, g:{}, sg:{})",
            ctx.access_key,
            params.group,
            params.scaling_group,
        )
        result = await self._resource_preset.check_presets.wait_for_complete(
            CheckResourcePresetsAction(
                access_key=AccessKey(ctx.access_key),
                resource_policy=resource_policy,
                domain_name=ctx.user_domain,
                user_id=ctx.user_uuid,
                group=params.group,
                scaling_group=params.scaling_group,
            )
        )
        scaling_groups_json: dict[str, Any] = {}
        for sgname, sg_data in result.scaling_groups.items():
            scaling_groups_json[sgname] = {
                ResourceSlotState.OCCUPIED: quantities_to_json(sg_data[ResourceSlotState.OCCUPIED]),
                ResourceSlotState.AVAILABLE: quantities_to_json(
                    sg_data[ResourceSlotState.AVAILABLE]
                ),
            }
        resp = CheckPresetsResponse(
            presets=result.presets,
            keypair_limits=quantities_to_json(result.keypair_limits),
            keypair_using=quantities_to_json(result.keypair_using),
            keypair_remaining=quantities_to_json(result.keypair_remaining),
            group_limits=quantities_to_json(result.group_limits),
            group_using=quantities_to_json(result.group_using),
            group_remaining=quantities_to_json(result.group_remaining),
            scaling_group_remaining=quantities_to_json(result.scaling_group_remaining),
            scaling_groups=scaling_groups_json,
        )
        return APIResponse.build(HTTPStatus.OK, resp)

    # ------------------------------------------------------------------
    # recalculate_usage (POST /resource/recalculate-usage)
    # ------------------------------------------------------------------

    async def recalculate_usage(self, ctx: UserContext) -> APIResponse:
        log.info("RECALCULATE_USAGE ()")
        await self._agent.recalculate_usage.wait_for_complete(RecalculateUsageAction())
        return APIResponse.build(HTTPStatus.OK, EmptyResponse())

    # ------------------------------------------------------------------
    # usage_per_month (GET /resource/usage/month)
    # ------------------------------------------------------------------

    async def usage_per_month(
        self,
        query: QueryParam[UsagePerMonthQuery],
        ctx: UserContext,
    ) -> APIResponse:
        params = query.parsed
        log.info(
            "USAGE_PER_MONTH (g:[{}], month:{})",
            ",".join(str(gid) for gid in params.group_ids) if params.group_ids else "",
            params.month,
        )
        result = await self._group.usage_per_month.wait_for_complete(
            UsagePerMonthAction(
                group_ids=params.group_ids,
                month=params.month,
            )
        )
        return APIResponse.build(HTTPStatus.OK, RawListResponse(root=result.result))

    # ------------------------------------------------------------------
    # usage_per_period (GET /resource/usage/period)
    # ------------------------------------------------------------------

    async def usage_per_period(
        self,
        query: QueryParam[UsagePerPeriodQuery],
        ctx: UserContext,
    ) -> APIResponse:
        params = query.parsed
        result = await self._group.usage_per_period.wait_for_complete(
            UsagePerPeriodAction(
                project_id=params.project_id,
                start_date=params.start_date,
                end_date=params.end_date,
            )
        )
        return APIResponse.build(HTTPStatus.OK, RawListResponse(root=result.result))

    # ------------------------------------------------------------------
    # user_month_stats (GET /resource/stats/user/month)
    # ------------------------------------------------------------------

    async def user_month_stats(self, ctx: UserContext) -> APIResponse:
        log.info("USER_LAST_MONTH_STATS (ak:{}, u:{})", ctx.access_key, ctx.user_uuid)
        result = await self._user.user_month_stats.wait_for_complete(
            UserMonthStatsAction(user_id=ctx.user_uuid)
        )
        return APIResponse.build(HTTPStatus.OK, RawListResponse(root=result.stats))

    # ------------------------------------------------------------------
    # admin_month_stats (GET /resource/stats/admin/month)
    # ------------------------------------------------------------------

    async def admin_month_stats(self, ctx: UserContext) -> APIResponse:
        log.info("ADMIN_LAST_MONTH_STATS ()")
        result = await self._user.admin_month_stats.wait_for_complete(AdminMonthStatsAction())
        return APIResponse.build(HTTPStatus.OK, RawListResponse(root=result.stats))

    # ------------------------------------------------------------------
    # get_watcher_status (GET /resource/watcher)
    # ------------------------------------------------------------------

    async def get_watcher_status(
        self,
        query: QueryParam[WatcherStatusQuery],
        ctx: UserContext,
    ) -> APIResponse:
        params = query.parsed
        log.info("GET_WATCHER_STATUS (ag:{})", params.agent_id)
        result = await self._agent.get_watcher_status.wait_for_complete(
            GetWatcherStatusAction(agent_id=AgentId(params.agent_id))
        )
        return APIResponse.build(HTTPStatus.OK, WatcherDataResponse(root=result.data))

    # ------------------------------------------------------------------
    # watcher_agent_start (POST /resource/watcher/agent/start)
    # ------------------------------------------------------------------

    async def watcher_agent_start(
        self,
        body: BodyParam[WatcherAgentRequest],
        ctx: UserContext,
    ) -> APIResponse:
        params = body.parsed
        log.info("WATCHER_AGENT_START (ag:{})", params.agent_id)
        result = await self._agent.watcher_agent_start.wait_for_complete(
            WatcherAgentStartAction(agent_id=AgentId(params.agent_id))
        )
        return APIResponse.build(HTTPStatus.OK, WatcherDataResponse(root=result.data))

    # ------------------------------------------------------------------
    # watcher_agent_stop (POST /resource/watcher/agent/stop)
    # ------------------------------------------------------------------

    async def watcher_agent_stop(
        self,
        body: BodyParam[WatcherAgentRequest],
        ctx: UserContext,
    ) -> APIResponse:
        params = body.parsed
        log.info("WATCHER_AGENT_STOP (ag:{})", params.agent_id)
        result = await self._agent.watcher_agent_stop.wait_for_complete(
            WatcherAgentStopAction(agent_id=AgentId(params.agent_id))
        )
        return APIResponse.build(HTTPStatus.OK, WatcherDataResponse(root=result.data))

    # ------------------------------------------------------------------
    # watcher_agent_restart (POST /resource/watcher/agent/restart)
    # ------------------------------------------------------------------

    async def watcher_agent_restart(
        self,
        body: BodyParam[WatcherAgentRequest],
        ctx: UserContext,
    ) -> APIResponse:
        params = body.parsed
        log.info("WATCHER_AGENT_RESTART (ag:{})", params.agent_id)
        result = await self._agent.watcher_agent_restart.wait_for_complete(
            WatcherAgentRestartAction(agent_id=AgentId(params.agent_id))
        )
        return APIResponse.build(HTTPStatus.OK, WatcherDataResponse(root=result.data))

    # ------------------------------------------------------------------
    # get_container_registries (GET /resource/container-registries)
    # ------------------------------------------------------------------

    async def get_container_registries(self, ctx: UserContext) -> APIResponse:
        result = await self._container_registry.get_container_registries.wait_for_complete(
            GetContainerRegistriesAction()
        )
        return APIResponse.build(HTTPStatus.OK, ContainerRegistriesResponse(root=result.registries))



from __future__ import annotations

from ai.backend.manager.data.resource_slot.types import (
    AgentResourceData,
    NumberFormatData,
    ResourceAllocationData,
    ResourceOccupancy,
    ResourceSlotTypeData,
)
from ai.backend.manager.repositories.resource_slot.repository import ResourceSlotRepository

from .actions.get_agent_resources import GetAgentResourcesAction, GetAgentResourcesResult
from .actions.get_domain_resource_overview import (
    GetDomainResourceOverviewAction,
    GetDomainResourceOverviewResult,
)
from .actions.get_kernel_allocations import GetKernelAllocationsAction, GetKernelAllocationsResult
from .actions.get_project_resource_overview import (
    GetProjectResourceOverviewAction,
    GetProjectResourceOverviewResult,
)
from .actions.get_resource_slot_type import GetResourceSlotTypeAction, GetResourceSlotTypeResult
from .actions.search_agent_resources import SearchAgentResourcesAction, SearchAgentResourcesResult
from .actions.search_resource_allocations import (
    SearchResourceAllocationsAction,
    SearchResourceAllocationsResult,
)
from .actions.search_resource_slot_types import (
    SearchResourceSlotTypesAction,
    SearchResourceSlotTypesResult,
)


class ResourceSlotService:
    _repository: ResourceSlotRepository

    def __init__(self, repository: ResourceSlotRepository) -> None:
        self._repository = repository

    async def get_agent_resources(self, action: GetAgentResourcesAction) -> GetAgentResourcesResult:
        rows = await self._repository.get_agent_resources(action.agent_id)
        items = [
            AgentResourceData(
                agent_id=row.agent_id,
                slot_name=row.slot_name,
                capacity=row.capacity,
                used=row.used,
            )
            for row in rows
        ]
        return GetAgentResourcesResult(items=items)

    async def search_agent_resources(
        self, action: SearchAgentResourcesAction
    ) -> SearchAgentResourcesResult:
        result = await self._repository.search_agent_resources(action.querier)
        return SearchAgentResourcesResult(
            items=result.items,
            total_count=result.total_count,
            has_next_page=result.has_next_page,
            has_previous_page=result.has_previous_page,
        )

    async def get_kernel_allocations(
        self, action: GetKernelAllocationsAction
    ) -> GetKernelAllocationsResult:
        rows = await self._repository.get_kernel_allocations(action.kernel_id)
        items = [
            ResourceAllocationData(
                kernel_id=row.kernel_id,
                slot_name=row.slot_name,
                requested=row.requested,
                used=row.used,
            )
            for row in rows
        ]
        return GetKernelAllocationsResult(items=items)

    async def search_resource_allocations(
        self, action: SearchResourceAllocationsAction
    ) -> SearchResourceAllocationsResult:
        result = await self._repository.search_resource_allocations(action.querier)
        return SearchResourceAllocationsResult(
            items=result.items,
            total_count=result.total_count,
            has_next_page=result.has_next_page,
            has_previous_page=result.has_previous_page,
        )

    async def get_resource_slot_type(
        self, action: GetResourceSlotTypeAction
    ) -> GetResourceSlotTypeResult:
        row = await self._repository.get_slot_type(action.slot_name)
        item = ResourceSlotTypeData(
            slot_name=row.slot_name,
            slot_type=row.slot_type,
            display_name=row.display_name,
            description=row.description,
            display_unit=row.display_unit,
            display_icon=row.display_icon,
            number_format=NumberFormatData(
                binary=row.number_format.binary,
                round_length=row.number_format.round_length,
            ),
            rank=row.rank,
        )
        return GetResourceSlotTypeResult(item=item)

    async def search_resource_slot_types(
        self, action: SearchResourceSlotTypesAction
    ) -> SearchResourceSlotTypesResult:
        result = await self._repository.search_slot_types(action.querier)
        return SearchResourceSlotTypesResult(
            items=result.items,
            total_count=result.total_count,
            has_next_page=result.has_next_page,
            has_previous_page=result.has_previous_page,
        )

    async def get_domain_resource_overview(
        self, action: GetDomainResourceOverviewAction
    ) -> GetDomainResourceOverviewResult:
        occupancy: ResourceOccupancy = await self._repository.get_domain_resource_overview(
            action.domain_name
        )
        return GetDomainResourceOverviewResult(item=occupancy)

    async def get_project_resource_overview(
        self, action: GetProjectResourceOverviewAction
    ) -> GetProjectResourceOverviewResult:
        occupancy: ResourceOccupancy = await self._repository.get_project_resource_overview(
            action.project_id
        )
        return GetProjectResourceOverviewResult(item=occupancy)



"""Deployment strategy evaluation for rolling update and blue-green deployments (BEP-1049)."""

from .blue_green import BlueGreenStrategy
from .rolling_update import RollingUpdateStrategy
from .types import AbstractDeploymentStrategy

__all__ = [
    "AbstractDeploymentStrategy",
    "BlueGreenStrategy",
    "RollingUpdateStrategy",
]



"""Scaling group module registrar."""

from __future__ import annotations

from typing import TYPE_CHECKING

from ai.backend.manager.api.rest.middleware.auth import auth_required
from ai.backend.manager.api.rest.routing import RouteRegistry

from .handler import ScalingGroupHandler

if TYPE_CHECKING:
    from ai.backend.manager.api.rest.types import RouteDeps


def register_scaling_group_routes(
    handler: ScalingGroupHandler, route_deps: RouteDeps
) -> RouteRegistry:
    """Build the scaling group sub-application."""

    reg = RouteRegistry.create("scaling-groups", route_deps.cors_options)

    reg.add(
        "GET",
        "",
        handler.list_available_sgroups,
        middlewares=[auth_required, route_deps.read_status_mw],
    )
    reg.add(
        "GET",
        "/{scaling_group}/wsproxy-version",
        handler.get_wsproxy_version,
        middlewares=[auth_required, route_deps.read_status_mw],
    )
    return reg



from __future__ import annotations

import enum
import uuid
from dataclasses import dataclass


class DotfileScope(enum.StrEnum):
    DOMAIN = "domain"
    GROUP = "group"
    USER = "user"


DotfileEntityKey = str | uuid.UUID


@dataclass(frozen=True)
class DotfileEntry:
    path: str
    perm: str
    data: str


@dataclass(frozen=True)
class DotfileQueryResult:
    entries: list[DotfileEntry]
    leftover_space: int



from __future__ import annotations

import uuid
from collections.abc import Mapping
from dataclasses import dataclass, field
from typing import Any, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.models.vfolder import VFolderOperationStatus, VFolderPermission

from .base import VFolderAction


@dataclass
class VFolderSharedInfo:
    """Domain type for shared vfolder permission info."""

    vfolder_id: uuid.UUID
    vfolder_name: str
    status: VFolderOperationStatus
    owner: str
    folder_type: str  # "project" or "user"
    shared_user_uuid: uuid.UUID
    shared_user_email: str
    permission: VFolderPermission


@dataclass
class ShareVFolderAction(VFolderAction):
    """Share a group vfolder with users by granting permissions directly."""

    user_uuid: uuid.UUID
    vfolder_uuid: uuid.UUID
    resource_policy: Mapping[str, Any]
    permission: VFolderPermission
    emails: list[str] = field(default_factory=list)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE

    @override
    def entity_id(self) -> str | None:
        return str(self.vfolder_uuid)


@dataclass
class ShareVFolderActionResult(BaseActionResult):
    shared_emails: list[str] = field(default_factory=list)

    @override
    def entity_id(self) -> str | None:
        return None


@dataclass
class UnshareVFolderAction(VFolderAction):
    """Revoke direct sharing permissions from users."""

    user_uuid: uuid.UUID
    vfolder_uuid: uuid.UUID
    resource_policy: Mapping[str, Any]
    emails: list[str] = field(default_factory=list)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.DELETE

    @override
    def entity_id(self) -> str | None:
        return str(self.vfolder_uuid)


@dataclass
class UnshareVFolderActionResult(BaseActionResult):
    unshared_emails: list[str] = field(default_factory=list)

    @override
    def entity_id(self) -> str | None:
        return None


@dataclass
class ListSharedVFoldersAction(VFolderAction):
    """List all shared vfolder permissions, optionally filtered by vfolder ID."""

    vfolder_id: uuid.UUID | None = None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH

    @override
    def entity_id(self) -> str | None:
        return str(self.vfolder_id) if self.vfolder_id else None


@dataclass
class ListSharedVFoldersActionResult(BaseActionResult):
    shared: list[VFolderSharedInfo] = field(default_factory=list)

    @override
    def entity_id(self) -> str | None:
        return None


@dataclass
class UpdateVFolderSharingStatusAction(VFolderAction):
    """Batch update or delete sharing permissions."""

    vfolder_id: uuid.UUID = field(default_factory=uuid.uuid4)
    to_delete: list[uuid.UUID] = field(default_factory=list)
    to_update: list[tuple[uuid.UUID, VFolderPermission]] = field(default_factory=list)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE

    @override
    def entity_id(self) -> str | None:
        return str(self.vfolder_id)


@dataclass
class UpdateVFolderSharingStatusActionResult(BaseActionResult):
    @override
    def entity_id(self) -> str | None:
        return None



"""
Path parameter DTOs for domain REST API endpoints.
"""

from pydantic import Field

from ai.backend.common.api_handlers import BaseRequestModel


class GetDomainPathParam(BaseRequestModel):
    """Path parameter for getting a domain."""

    domain_name: str = Field(description="The domain name to retrieve")


class UpdateDomainPathParam(BaseRequestModel):
    """Path parameter for updating a domain."""

    domain_name: str = Field(description="The domain name to update")



"""Deployment module registrar."""

from __future__ import annotations

from typing import TYPE_CHECKING

from ai.backend.manager.api.rest.middleware.auth import auth_required
from ai.backend.manager.api.rest.routing import RouteRegistry

from .handler import DeploymentAPIHandler

if TYPE_CHECKING:
    from ai.backend.manager.api.rest.types import RouteDeps


def register_deployment_routes(
    handler: DeploymentAPIHandler, route_deps: RouteDeps
) -> RouteRegistry:
    """Build the deployment sub-application."""
    reg = RouteRegistry.create("deployments", route_deps.cors_options)

    # Deployment routes
    reg.add("POST", "", handler.create_deployment, middlewares=[auth_required])
    reg.add("POST", "/search", handler.search_deployments, middlewares=[auth_required])
    reg.add("GET", "/{deployment_id}", handler.get_deployment, middlewares=[auth_required])
    reg.add(
        "PATCH",
        "/{deployment_id}",
        handler.update_deployment,
        middlewares=[auth_required],
    )
    reg.add(
        "DELETE",
        "/{deployment_id}",
        handler.destroy_deployment,
        middlewares=[auth_required],
    )

    # Revision routes (nested under deployment)
    reg.add(
        "POST",
        "/{deployment_id}/revisions",
        handler.add_revision,
        middlewares=[auth_required],
    )
    reg.add(
        "POST",
        "/{deployment_id}/revisions/search",
        handler.search_revisions,
        middlewares=[auth_required],
    )
    reg.add(
        "GET",
        "/{deployment_id}/revisions/{revision_id}",
        handler.get_revision,
        middlewares=[auth_required],
    )
    reg.add(
        "POST",
        "/{deployment_id}/revisions/{revision_id}/activate",
        handler.activate_revision,
        middlewares=[auth_required],
    )
    reg.add(
        "POST",
        "/{deployment_id}/revisions/{revision_id}/deactivate",
        handler.deactivate_revision,
        middlewares=[auth_required],
    )

    # Policy routes (nested under deployment)
    reg.add(
        "PUT",
        "/{deployment_id}/policy",
        handler.upsert_deployment_policy,
        middlewares=[auth_required],
    )
    reg.add(
        "GET",
        "/{deployment_id}/policy",
        handler.get_deployment_policy,
        middlewares=[auth_required],
    )

    # Route routes (nested under deployment)
    reg.add(
        "POST",
        "/{deployment_id}/routes/search",
        handler.search_routes,
        middlewares=[auth_required],
    )
    reg.add(
        "PATCH",
        "/{deployment_id}/routes/{route_id}/traffic-status",
        handler.update_route_traffic_status,
        middlewares=[auth_required],
    )
    return reg



"""Query conditions and orders for image repository."""

from __future__ import annotations

import uuid
from collections.abc import Collection
from datetime import datetime

import sqlalchemy as sa

from ai.backend.common.data.filter_specs import StringMatchSpec
from ai.backend.common.types import ImageID
from ai.backend.manager.data.image.types import ImageStatus
from ai.backend.manager.models.image import ImageAliasRow, ImageRow
from ai.backend.manager.models.kernel.row import KernelRow
from ai.backend.manager.repositories.base import QueryCondition, QueryOrder


class ImageConditions:
    """Query conditions for images."""

    @staticmethod
    def by_ids(image_ids: Collection[ImageID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ImageRow.id.in_(image_ids)

        return inner

    @staticmethod
    def by_canonicals(canonicals: Collection[str]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ImageRow.name.in_(canonicals)

        return inner

    @staticmethod
    def by_statuses(statuses: Collection[ImageStatus]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ImageRow.status.in_(statuses)

        return inner

    @staticmethod
    def by_architecture(architecture: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ImageRow.architecture == architecture

        return inner

    @staticmethod
    def by_registry_id(registry_id: uuid.UUID) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ImageRow.registry_id == registry_id

        return inner

    # String filter factories for name
    @staticmethod
    def by_name_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ImageRow.name.ilike(f"%{spec.value}%")
            else:
                condition = ImageRow.name.like(f"%{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_name_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = sa.func.lower(ImageRow.name) == spec.value.lower()
            else:
                condition = ImageRow.name == spec.value
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_name_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ImageRow.name.ilike(f"{spec.value}%")
            else:
                condition = ImageRow.name.like(f"{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_name_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ImageRow.name.ilike(f"%{spec.value}")
            else:
                condition = ImageRow.name.like(f"%{spec.value}")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    # String filter factories for architecture
    @staticmethod
    def by_architecture_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ImageRow.architecture.ilike(f"%{spec.value}%")
            else:
                condition = ImageRow.architecture.like(f"%{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_architecture_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = sa.func.lower(ImageRow.architecture) == spec.value.lower()
            else:
                condition = ImageRow.architecture == spec.value
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_architecture_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ImageRow.architecture.ilike(f"{spec.value}%")
            else:
                condition = ImageRow.architecture.like(f"{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_architecture_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ImageRow.architecture.ilike(f"%{spec.value}")
            else:
                condition = ImageRow.architecture.like(f"%{spec.value}")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    # DateTime filter factories for last_used (computed from KernelRow)
    @staticmethod
    def _last_used_subquery() -> sa.ScalarSelect[datetime | None]:
        return (
            sa.select(sa.func.max(KernelRow.created_at))
            .where(
                sa.and_(
                    KernelRow.image == ImageRow.name,
                    KernelRow.architecture == ImageRow.architecture,
                )
            )
            .correlate(ImageRow)
            .scalar_subquery()
        )

    @staticmethod
    def by_last_used_before(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ImageConditions._last_used_subquery() < dt

        return inner

    @staticmethod
    def by_last_used_after(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ImageConditions._last_used_subquery() > dt

        return inner

    # String filter factories for alias (exists subquery-based)
    @staticmethod
    def by_alias_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subq = sa.select(sa.literal(1)).where(ImageAliasRow.image_id == ImageRow.id)
            if spec.case_insensitive:
                subq = subq.where(ImageAliasRow.alias.ilike(f"%{spec.value}%"))
            else:
                subq = subq.where(ImageAliasRow.alias.like(f"%{spec.value}%"))
            condition: sa.sql.expression.ColumnElement[bool] = sa.exists(subq)
            if spec.negated:
                condition = ~condition
            return condition

        return inner

    @staticmethod
    def by_alias_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subq = sa.select(sa.literal(1)).where(ImageAliasRow.image_id == ImageRow.id)
            if spec.case_insensitive:
                subq = subq.where(sa.func.lower(ImageAliasRow.alias) == spec.value.lower())
            else:
                subq = subq.where(ImageAliasRow.alias == spec.value)
            condition: sa.sql.expression.ColumnElement[bool] = sa.exists(subq)
            if spec.negated:
                condition = ~condition
            return condition

        return inner

    @staticmethod
    def by_alias_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subq = sa.select(sa.literal(1)).where(ImageAliasRow.image_id == ImageRow.id)
            if spec.case_insensitive:
                subq = subq.where(ImageAliasRow.alias.ilike(f"{spec.value}%"))
            else:
                subq = subq.where(ImageAliasRow.alias.like(f"{spec.value}%"))
            condition: sa.sql.expression.ColumnElement[bool] = sa.exists(subq)
            if spec.negated:
                condition = ~condition
            return condition

        return inner

    @staticmethod
    def by_alias_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subq = sa.select(sa.literal(1)).where(ImageAliasRow.image_id == ImageRow.id)
            if spec.case_insensitive:
                subq = subq.where(ImageAliasRow.alias.ilike(f"%{spec.value}"))
            else:
                subq = subq.where(ImageAliasRow.alias.like(f"%{spec.value}"))
            condition: sa.sql.expression.ColumnElement[bool] = sa.exists(subq)
            if spec.negated:
                condition = ~condition
            return condition

        return inner


class ImageOrders:
    """Query orders for images."""

    @staticmethod
    def name(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ImageRow.name.asc()
        return ImageRow.name.desc()

    @staticmethod
    def created_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ImageRow.created_at.asc()
        return ImageRow.created_at.desc()

    @staticmethod
    def architecture(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ImageRow.architecture.asc()
        return ImageRow.architecture.desc()

    @staticmethod
    def size_bytes(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ImageRow.size_bytes.asc()
        return ImageRow.size_bytes.desc()

    @staticmethod
    def registry(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ImageRow.registry.asc()
        return ImageRow.registry.desc()

    @staticmethod
    def tag(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ImageRow.tag.asc()
        return ImageRow.tag.desc()

    @staticmethod
    def status(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ImageRow.status.asc()
        return ImageRow.status.desc()

    @staticmethod
    def last_used(ascending: bool = True) -> QueryOrder:
        last_used_subq = ImageConditions._last_used_subquery()
        if ascending:
            return last_used_subq.asc()
        return last_used_subq.desc()


class ImageAliasConditions:
    """Query conditions for image aliases."""

    @staticmethod
    def by_ids(alias_ids: Collection[uuid.UUID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ImageAliasRow.id.in_(alias_ids)

        return inner

    @staticmethod
    def by_image_ids(image_ids: Collection[ImageID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ImageAliasRow.image_id.in_(image_ids)

        return inner

    # String filter factories for alias
    @staticmethod
    def by_alias_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ImageAliasRow.alias.ilike(f"%{spec.value}%")
            else:
                condition = ImageAliasRow.alias.like(f"%{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_alias_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = sa.func.lower(ImageAliasRow.alias) == spec.value.lower()
            else:
                condition = ImageAliasRow.alias == spec.value
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_alias_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ImageAliasRow.alias.ilike(f"{spec.value}%")
            else:
                condition = ImageAliasRow.alias.like(f"{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_alias_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = ImageAliasRow.alias.ilike(f"%{spec.value}")
            else:
                condition = ImageAliasRow.alias.like(f"%{spec.value}")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner


class ImageAliasOrders:
    """Query orders for image aliases."""

    @staticmethod
    def alias(ascending: bool = True) -> QueryOrder:
        if ascending:
            return ImageAliasRow.alias.asc()
        return ImageAliasRow.alias.desc()



from pydantic import Field

from ai.backend.common.api_handlers import BaseRequestModel


class GetQuotaScopePathParam(BaseRequestModel):
    storage_host_name: str = Field(description="The storage host name")
    quota_scope_id: str = Field(
        description="The quota scope ID (e.g. 'user:<uuid>' or 'project:<uuid>')"
    )



from .fetch_status import FetchManagerStatusAction, FetchManagerStatusActionResult
from .get_announcement import GetAnnouncementAction, GetAnnouncementActionResult
from .get_db_cxn_status import GetDbCxnStatusAction, GetDbCxnStatusActionResult
from .perform_scheduler_ops import PerformSchedulerOpsAction, PerformSchedulerOpsActionResult
from .update_announcement import UpdateAnnouncementAction, UpdateAnnouncementActionResult
from .update_status import UpdateManagerStatusAction, UpdateManagerStatusActionResult

__all__ = (
    "FetchManagerStatusAction",
    "FetchManagerStatusActionResult",
    "GetAnnouncementAction",
    "GetAnnouncementActionResult",
    "GetDbCxnStatusAction",
    "GetDbCxnStatusActionResult",
    "PerformSchedulerOpsAction",
    "PerformSchedulerOpsActionResult",
    "UpdateAnnouncementAction",
    "UpdateAnnouncementActionResult",
    "UpdateManagerStatusAction",
    "UpdateManagerStatusActionResult",
)



from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.common.types import KernelId
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.services.stream.actions.base import StreamAction


@dataclass(frozen=True)
class GCStaleConnectionsAction(StreamAction):
    active_session_ids: list[KernelId]

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.DELETE


@dataclass(frozen=True)
class GCStaleConnectionsActionResult(BaseActionResult):
    removed_sessions: list[str]

    @override
    def entity_id(self) -> str | None:
        return None



"""Stream module registrar."""

from __future__ import annotations

from typing import TYPE_CHECKING

from ai.backend.manager.api.rest.middleware.auth import auth_required
from ai.backend.manager.api.rest.routing import RouteRegistry

from .handler import StreamHandler

if TYPE_CHECKING:
    from ai.backend.manager.api.rest.types import RouteDeps
    from ai.backend.manager.event_dispatcher.handlers.stream_cleanup import (
        StreamCleanupEventHandler,
    )
    from ai.backend.manager.services.stream.processors import StreamProcessors


def register_stream_routes(
    handler: StreamHandler,
    route_deps: RouteDeps,
    *,
    stream_processors: StreamProcessors,
    stream_cleanup_handler: StreamCleanupEventHandler,
) -> RouteRegistry:
    """Build the stream sub-application."""
    from .handler import PrivateContext, stream_app_ctx, stream_shutdown

    reg = RouteRegistry.create("stream", route_deps.cors_options)
    ctx = PrivateContext()

    # Wire lifecycle hooks — capture deps via closure
    reg.app.cleanup_ctx.append(
        lambda app: stream_app_ctx(
            app,
            ctx,
            stream_processors=stream_processors,
            stream_cleanup_handler=stream_cleanup_handler,
        )
    )
    reg.app.on_shutdown.append(lambda app: stream_shutdown(app, ctx))

    _mw = [route_deps.read_status_mw, auth_required]

    reg.add("GET", r"/session/{session_name}/pty", handler.stream_pty, middlewares=_mw)
    reg.add("GET", r"/session/{session_name}/execute", handler.stream_execute, middlewares=_mw)
    reg.add("GET", r"/session/{session_name}/apps", handler.get_stream_apps, middlewares=_mw)
    reg.add("GET", r"/session/{session_name}/httpproxy", handler.stream_proxy, middlewares=_mw)
    reg.add("GET", r"/session/{session_name}/tcpproxy", handler.stream_proxy, middlewares=_mw)
    return reg



from dataclasses import dataclass
from typing import Any, override
from uuid import UUID

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.services.group.actions.base import GroupAction


# TODO: Change to batch action
@dataclass
class UsagePerPeriodAction(GroupAction):
    start_date: str
    end_date: str
    project_id: UUID | None = None

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH


@dataclass
class UsagePerPeriodActionResult(BaseActionResult):
    # TODO: Define return type
    result: list[Any]

    @override
    def entity_id(self) -> str | None:
        return None
